/**
 * @file
 * VuoRunner implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code is proprietary, and may not be redistributed.
 * For more information, see http://vuo.org/license.
 */

#include "VuoRunner.hh"
#include "VuoFileUtilities.hh"

#include "VuoHeap.h"

#include <pthread.h>
#include <stdio.h>
#include <dlfcn.h>
#include <sstream>
#include <CoreFoundation/CoreFoundation.h>

static pthread_t mainThread = NULL;
static void __attribute__((constructor)) thisFunctionIsCalledAtStartup()
{
	mainThread = pthread_self();
}
static bool isMainThread(void)
{
	return mainThread == pthread_self();
}

/**
 * Creates a runner object that can run a composition in a new process.
 *
 * @param executablePath A linked composition executable, produced by @c VuoCompiler::linkCompositionToCreateExecutable().
 * @param deleteExecutableWhenFinished True if the runner should delete @c executablePath when it's finished using the file.
 *
 * @see CompileAndRunInNewProcess.cc
 */
VuoRunner * VuoRunner::newSeparateProcessRunnerFromExecutable(string executablePath, bool deleteExecutableWhenFinished)
{
	VuoRunner * vr = new VuoRunner();
	vr->executablePath = executablePath;
	vr->shouldDeleteBinariesWhenFinished = deleteExecutableWhenFinished;
	return vr;
}

/**
 * Creates a runner object that can run a composition in a new process
 * and replace the composition with a new version while it's running.
 *
 * @param compositionLoaderPath The VuoCompositionLoader executable.
 * @param compositionDylibPath A linked composition dynamic library, produced by @c VuoCompiler::linkCompositionToCreateDynamicLibraries().
 * @param resourceDylibPath A linked resource dynamic library, produced by the first call to @c VuoCompiler::linkCompositionToCreateDynamicLibraries() for this runner.
 * @param deleteDylibsWhenFinished True if the runner should delete @c compositionDylibPath and @c resourceDylibPath when it's finished using the files.
 */
VuoRunner * VuoRunner::newSeparateProcessRunnerFromDynamicLibrary(string compositionLoaderPath, string compositionDylibPath,
																  string resourceDylibPath, bool deleteDylibsWhenFinished)
{
	VuoRunner * vr = new VuoRunner();
	vr->executablePath = compositionLoaderPath;
	vr->dylibPath = compositionDylibPath;
	vr->resourceDylibPaths.push_back(resourceDylibPath);
	vr->shouldDeleteBinariesWhenFinished = deleteDylibsWhenFinished;
	return vr;
}

/**
 * Creates a runner object that can run a composition in the current process.
 *
 * @param dylibPath A linked composition dynamic library, produced by @c VuoCompiler::linkCompositionToCreateDynamicLibrary().
 * @param deleteDylibWhenFinished True if the runner should delete @c dylibPath when it's finished using the file.
 *
 * @see CompileAndRunInCurrentProcess.cc
 */
VuoRunner * VuoRunner::newCurrentProcessRunnerFromDynamicLibrary(string dylibPath, bool deleteDylibWhenFinished)
{
	VuoRunner * vr = new VuoRunner();
	vr->dylibPath = dylibPath;
	vr->shouldDeleteBinariesWhenFinished = deleteDylibWhenFinished;
	return vr;
}

/**
 * Destructor.
 *
 * Assumes the composition either has not been started or has been started and stopped.
 */
VuoRunner::~VuoRunner(void)
{
	dispatch_release(stoppedSemaphore);
	dispatch_release(beganListeningSemaphore);
	dispatch_release(endedListeningSemaphore);
	dispatch_release(delegateQueue);
	dispatch_release(anyPublishedPortEventSemaphore);
}

/**
 * Private constructor, used by factory methods.
 */
VuoRunner::VuoRunner(void)
{
	dylibHandle = NULL;
	shouldDeleteBinariesWhenFinished = false;
	stopped = true;
	listenCanceled = false;
	stoppedSemaphore = dispatch_semaphore_create(1);
	beganListeningSemaphore = dispatch_semaphore_create(0);
	endedListeningSemaphore = dispatch_semaphore_create(1);
	controlQueue = dispatch_queue_create("org.vuo.runner.control", NULL);
	ZMQContext = NULL;
	ZMQControl = NULL;
	ZMQTelemetry = NULL;
	ZMQLoaderControl = NULL;
	delegate = NULL;
	delegateQueue = dispatch_queue_create("org.vuo.runner.delegate", NULL);
	arePublishedInputPortsCached = false;
	arePublishedOutputPortsCached = false;
	anyPublishedPortEventSemaphore = dispatch_semaphore_create(0);
	anyPublishedPortEventSignaled = false;

	// Processes using VuoRunner are also likely to use Vuo Types, which need reference counting support
	initializeReferenceCounts();
}

/**
 * Starts the composition running.
 *
 * If running the composition in the current process, a call to this method must be followed by
 * either a call to @c runOnMainThread() or repeated calls to @c drainMainDispatchQueue()
 * in order to run the composition.
 *
 * If running the composition in a separate process, no further calls are needed.
 *
 * Assumes the composition has not been started.
 */
void VuoRunner::start(void)
{
	start(false);
}

/**
 * Starts the composition running, but in a paused state.
 * Useful in conjunction with @c setPublishedInputPortValue().
 *
 * If running the composition in the current process, a call to this method must be followed by
 * either a call to @c runOnMainThread() or repeated calls to @c drainMainDispatchQueue()
 * in order to run the composition.
 *
 * If running the composition in a separate process, no further calls are needed.
 *
 * Assumes the composition has not been started.
 */
void VuoRunner::startPaused(void)
{
	start(true);
}

/**
 * Internal helper method for starting the composition running.
 *
 * Assumes the composition has not been started.
 *
 * @param isPaused If true, the composition is initially in a paused state.
 */
void VuoRunner::start(bool isPaused)
{
	stopped = false;
	dispatch_semaphore_wait(stoppedSemaphore, DISPATCH_TIME_FOREVER);

	ZMQContext = zmq_init(1);

	if (isInCurrentProcess())
	{
		// Start the composition in the current process.

		dylibHandle = dlopen(dylibPath.c_str(), RTLD_NOW);
		if (!dylibHandle)
		{
			fprintf(stderr, "VuoRunner: couldn't load dylib: %s\n", dlerror());
			return;
		}

		typedef void(* initInProcessType)(void *_ZMQContext, const char *controlURL, const char *telemetryURL, bool _isPaused);

		initInProcessType vuoInitInProcess = (initInProcessType)dlsym(dylibHandle, "vuoInitInProcess");
		if (! vuoInitInProcess)
		{
			fprintf(stderr, "VuoRunner: couldn't find function `vuoInitInProcess`: %s\n", dlerror());
			return;
		}

		ZMQControlURL = "inproc://" + VuoFileUtilities::makeTmpFile("vuo-control", "");
		ZMQTelemetryURL = "inproc://" + VuoFileUtilities::makeTmpFile("vuo-telemetry", "");

		vuoInitInProcess(ZMQContext, ZMQControlURL.c_str(), ZMQTelemetryURL.c_str(), isPaused);
	}
	else
	{
		// Start the composition or composition loader in a new process.

		vector<string> args;

		string executableName;
		if (isUsingCompositionLoader())
		{
			// If we're using the loader, set the executable's display name to the dylib,
			// so that composition's name shows up in the process list.
			string dir, file, ext;
			VuoFileUtilities::splitPath(dylibPath, dir, file, ext);
			executableName = file;
		}
		else
		{
			string dir, file, ext;
			VuoFileUtilities::splitPath(executablePath, dir, file, ext);
			string executableName = file;
			if (! ext.empty())
				executableName += "." + ext;
		}
		args.push_back(executableName);

		ZMQControlURL = "ipc://" + VuoFileUtilities::makeTmpFile("vuo-control", "");
		ZMQTelemetryURL = "ipc://" + VuoFileUtilities::makeTmpFile("vuo-telemetry", "");
		args.push_back("--vuo-control=" + ZMQControlURL);
		args.push_back("--vuo-telemetry=" + ZMQTelemetryURL);

		if (isUsingCompositionLoader())
		{
			ZMQLoaderControlURL = "ipc://" + VuoFileUtilities::makeTmpFile("vuo-loader", "");
			args.push_back("--vuo-loader=" + ZMQLoaderControlURL);
		}

		if (isPaused)
		{
			args.push_back("--vuo-pause");
		}

		char * argv[7];
		for (int i = 0; i < args.size(); ++i)
		{
			argv[i] = (char *) malloc(args[i].length() + 1);
			strcpy(argv[i], args[i].c_str());
		}
		argv[args.size()] = NULL;

		compositionPid = fork();
		if (! compositionPid)
		{
			execv(executablePath.c_str(), argv);
			return;
		}

#ifdef MAC
		if (! fork())
		{
			// Bring the composition's window to the foreground.

			ostringstream setPid;
			setPid << "set pid to " << compositionPid;

			// osascript is codesigned, so we can't override its dylibs.
			// Unset the override, so we don't get a "dyld: DYLD_ environment variables being ignored because main executable (/usr/bin/osascript) is code signed with entitlements" warning.
			unsetenv("DYLD_LIBRARY_PATH");

			execl(
						"/usr/bin/osascript",
						"osascript",
						"-e", "tell application \"System Events\"",
							"-e", setPid.str().c_str(),
							"-e", "repeat 50 times",
								"-e", "try",
									"-e", "set (frontmost of processes whose unix id is pid) to true",
									"-e", "if (frontmost of processes whose unix id is pid) then exit repeat",
								"-e", "end try",
								"-e", "delay 0.1",
							"-e", "end repeat",
						"-e", "end tell",
						NULL
						);
		}
#endif
	}

	// Connect to the composition loader (if any) and composition.
	if (isUsingCompositionLoader())
	{
		ZMQLoaderControl = zmq_socket(ZMQContext,ZMQ_REQ);

		// Try to connect to the composition loader. If at first we don't succeed, give the composition loader a little more time to set up the socket.
		int numTries = 0;
		while (zmq_connect(ZMQLoaderControl,ZMQLoaderControlURL.c_str()))
		{
			if (++numTries == 1000)
			{
				fprintf(stderr, "runner VuoLoaderControl: connect failed.\n");
				return;
			}
			usleep(USEC_PER_SEC / 1000);
		}

		replaceComposition(dylibPath, resourceDylibPaths.at(0), "");
	}
	else
	{
		setUpConnections();
	}
}

/**
 * Establishes ZMQ connections with the running composition.
 */
void VuoRunner::setUpConnections(void)
{
	ZMQControl = zmq_socket(ZMQContext,ZMQ_REQ);

	// Try to connect to the composition. If at first we don't succeed, give the composition a little more time to set up the socket.
	int numTries = 0;
	while (zmq_connect(ZMQControl,ZMQControlURL.c_str()))
	{
		if (++numTries == 1000)
		{
			fprintf(stderr, "runner VuoControl: connect failed.\n");
			return;
		}
		usleep(USEC_PER_SEC / 1000);
	}

	// Cache published ports so they're available whenever a caller starts listening for published port value changes.
	if (isInCurrentProcess())
	{
		dispatch_queue_t queue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0);
		dispatch_async(queue, ^{
						   getCachedPublishedPorts(false);
						   getCachedPublishedPorts(true);
					   });
		while (! arePublishedOutputPortsCached)
		{
			drainMainDispatchQueue();
			usleep(USEC_PER_SEC / 10);
		}
	}
	else
	{
		getCachedPublishedPorts(false);
		getCachedPublishedPorts(true);
	}

	listenCanceled = false;
	dispatch_semaphore_wait(endedListeningSemaphore, DISPATCH_TIME_FOREVER);
	dispatch_queue_t queue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0);
	dispatch_async(queue, ^{
					   listen();
				   });
	dispatch_semaphore_wait(beganListeningSemaphore, DISPATCH_TIME_FOREVER);
}

/**
 * Private API function in libdispatch.
 */
extern "C"
{
extern void _dispatch_main_queue_callback_4CF(mach_msg_header_t *msg);
}

/**
 * For a composition in the current process,
 * runs the composition until it stops (either on its own or from a call to @c stop() on another thread).
 *
 * Call this function after @c start() or @c startPaused().
 *
 * There's no need to call this function if your application calls @c dispatch_main(), @c NSApplicationMain(), or @c UIApplicationMain(), or invokes a @c CFRunLoop on the main thread.
 *
 * Throws a @c std::logic_error if this @c VuoRunner was constructed with @c newSeparateProcessRunnerFromExecutable() or if this function is not called on the main thread.
 *
 * @see drainMainDispatchQueue(), an alternative to this method.
 */
void VuoRunner::runOnMainThread(void)
{
	if (! isInCurrentProcess())
		throw std::logic_error("The composition is not running in the current process. Only use this function if the composition was constructed with newCurrentProcessRunnerFromDynamicLibrary().");

	if (! isMainThread())
		throw std::logic_error("This is not the main thread. Only call this function from the main thread.");

	while (! stopped)
	{
		_dispatch_main_queue_callback_4CF(0);
		usleep(10000);
	}
}

/**
 * For a composition in the current process, briefly performs work that requires the main thread (such as Cocoa event handling).
 * Repeated calls to this function allow the composition to run without taking over the main thread like @c runOnMainThread() does.
 *
 * Start calling this function after @c start() or @c startPaused().
 *
 * There's no need to call this method if your application calls @c dispatch_main(), @c NSApplicationMain(), or @c UIApplicationMain(), or invokes a @c CFRunLoop on the main thread.
 *
 * Throws a @c std::logic_error if this @c VuoRunner was constructed with @c newSeparateProcessRunnerFromExecutable() or if this function is not called on the main thread.
 *
 * \eg{
 *   VuoRunner *runner = newCurrentProcessRunnerFromDynamicLibrary(...);
 *   runner->start();
 *   while (! runner->isStopped())
 *   {
 *      runner->drainMainDispatchQueue();
 *		// do other work on the main thread
 *   }
 * }
 *
 * @see runOnMainThread(), an alternative to this method.
 */
void VuoRunner::drainMainDispatchQueue(void)
{
	if (! isInCurrentProcess())
		throw std::logic_error("The composition is not running in the current process. Only use this function if the composition was constructed with newCurrentProcessRunnerFromDynamicLibrary().");

	if (! isMainThread())
		throw std::logic_error("This is not the main thread. Only call this function from the main thread.");

	_dispatch_main_queue_callback_4CF(0);
}

/**
 * Sends a control request to the composition telling it to cease triggering events.
 * When this method returns, no more events will be triggered and all events will have finished propagating through the composition.
 *
 * Assumes the composition has been started, is not paused, and has not been hibernated or stopped.
 */
void VuoRunner::pause(void)
{
	dispatch_sync(controlQueue, ^{
					  vuoMemoryBarrier();

					  vuoControlRequestSend(VuoControlRequestCompositionPause,NULL,0);
					  vuoControlReplyReceive(VuoControlReplyCompositionPaused);
				  });
}

/**
 * Sends a control request to the composition telling it to resume generating events.
 *
 * Assumes the composition is paused.
 */
void VuoRunner::unpause(void)
{
	dispatch_sync(controlQueue, ^{
					  vuoMemoryBarrier();

					  vuoControlRequestSend(VuoControlRequestCompositionUnpause,NULL,0);
					  vuoControlReplyReceive(VuoControlReplyCompositionUnpaused);
				  });
}

/**
 * Sends a control request to the composition loader telling it to load an updated version
 * of the running composition. When this method returns, the new composition will have started.
 *
 * Assumes the composition loader has been started and has not been stopped.
 *
 * @param compositionDylibPath A linked composition dynamic library, produced by @c VuoCompiler::linkCompositionToCreateDynamicLibrary().
 * @param resourceDylibPath A linked resource dynamic library, produced by a call to @c VuoCompiler::linkCompositionToCreateDynamicLibraries() for this runner.
 *				Pass an empty string if no linked resource dynamic library was created.
 * @param compositionDiff A comparison of the old and new compositions, produced by @c VuoCompilerComposition::diffAgainstOlderComposition().
 */
void VuoRunner::replaceComposition(string compositionDylibPath, string resourceDylibPath, string compositionDiff)
{
	if (! isUsingCompositionLoader())
		throw std::logic_error("The runner is not using a composition loader. Only use this function if the composition was constructed with newSeparateProcessRunnerFromDynamicLibrary().");

	dispatch_sync(controlQueue, ^{
					  if (dylibPath != compositionDylibPath)
					  {
						  if (shouldDeleteBinariesWhenFinished)
						  {
							  remove(dylibPath.c_str());
						  }

						  dylibPath = compositionDylibPath;
					  }

					  vuoMemoryBarrier();

					  cleanUpConnections();

					  zmq_msg_t messages[3];
					  vuoInitMessageWithString(&messages[0], dylibPath.c_str());
					  vuoInitMessageWithString(&messages[1], resourceDylibPath.c_str());
					  vuoInitMessageWithString(&messages[2], compositionDiff.c_str());

					  vuoLoaderControlRequestSend(VuoLoaderControlRequestCompositionReplace,messages,3);
					  vuoLoaderControlReplyReceive(VuoLoaderControlReplyCompositionReplaced);

					  setUpConnections();
				  });
}

/**
 * Sends a control request to the composition telling it to stop. When this method returns,
 * the composition process will be on its way to terminating.
 *
 * This method waits for any pending delegate methods to return.
 *
 * Assumes the composition has been started and has not been stopped.
 */
void VuoRunner::stop(void)
{
	dispatch_sync(controlQueue, ^{
					  vuoMemoryBarrier();

					  vuoControlRequestSend(VuoControlRequestCompositionStop,NULL,0);
					  vuoControlReplyReceive(VuoControlReplyCompositionStopping);

					  cleanUpConnections();

					  if (isUsingCompositionLoader())
					  {
						  zmq_close(ZMQLoaderControl);
						  ZMQLoaderControl = NULL;
					  }

					  if (isInCurrentProcess())
					  {
						  typedef void(* finiType)(void);

						  finiType vuoFini = (finiType)dlsym(dylibHandle, "vuoFini");
						  if (! vuoFini)
						  {
							  fprintf(stderr, "VuoRunner: couldn't find function `vuoFini`: %s\n", dlerror());
							  return;
						  }
						  vuoFini();

						  dlclose(dylibHandle);
						  dylibHandle = NULL;
					  }
					  else
					  {
						  // Wait for child process to end.
						  int status;
						  waitpid(compositionPid, &status, 0);

						  zmq_term(ZMQContext);
					  }

					  if (shouldDeleteBinariesWhenFinished)
					  {
						  if (isUsingCompositionLoader())
						  {
							  remove(dylibPath.c_str());
							  for (vector<string>::iterator i = resourceDylibPaths.begin(); i != resourceDylibPaths.end(); ++i)
								  remove((*i).c_str());
						  }
						  else if (isInCurrentProcess())
						  {
							  remove(dylibPath.c_str());
						  }
						  else
						  {
							  remove(executablePath.c_str());
						  }
					  }

					  stopped = true;
					  dispatch_semaphore_signal(stoppedSemaphore);
				  });
}

/**
 * Closes ZMQ connections to the running composition.
 */
void VuoRunner::cleanUpConnections(void)
{
	if (! ZMQControl)
		return;

	zmq_close(ZMQControl);
	ZMQControl = NULL;

	listenCanceled = true;
	dispatch_semaphore_wait(endedListeningSemaphore, DISPATCH_TIME_FOREVER);
	dispatch_semaphore_signal(endedListeningSemaphore);
}

/**
 * Waits until the composition is stopped.
 */
void VuoRunner::waitUntilStopped(void)
{
	dispatch_retain(stoppedSemaphore);
	dispatch_semaphore_wait(stoppedSemaphore, DISPATCH_TIME_FOREVER);
	dispatch_semaphore_signal(stoppedSemaphore);
	dispatch_release(stoppedSemaphore);
}

/**
 * Sends a control request to the composition telling it to modify an input port's value.
 * When this method returns, the input port value will have been modified.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 *
 * @param portIdentifier The compile-time identifier for the port (see @c VuoCompilerEventPort::getIdentifier()).
 * @param valueAsString String representation of the port's new value.
 *
 * @see VuoTypes for information about types and serialization.
 */
void VuoRunner::setInputPortValue(string portIdentifier, string valueAsString)
{
	dispatch_sync(controlQueue, ^{
					  vuoMemoryBarrier();

					  zmq_msg_t messages[2];
					  vuoInitMessageWithString(&messages[0], portIdentifier.c_str());
					  vuoInitMessageWithString(&messages[1], valueAsString.c_str());
					  vuoControlRequestSend(VuoControlRequestInputPortValueModify, messages, 2);
					  vuoControlReplyReceive(VuoControlReplyInputPortValueModified);
				  });
}

/**
 * Sends a control request to the composition to get an input port's value.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 *
 * @param portIdentifier The compile-time identifier for the port (see @c VuoCompilerEventPort::getIdentifier()).
 * @return String representation of the port's value.
 *
 * @see VuoTypes for information about types and serialization.
 */
string VuoRunner::getInputPortValue(string portIdentifier)
{
	__block string valueAsString;
	dispatch_sync(controlQueue, ^{
					  vuoMemoryBarrier();

					  zmq_msg_t messages[2];
					  vuoInitMessageWithBool(&messages[0], !isInCurrentProcess());
					  vuoInitMessageWithString(&messages[1], portIdentifier.c_str());
					  vuoControlRequestSend(VuoControlRequestInputPortValueRetrieve, messages, 2);
					  vuoControlReplyReceive(VuoControlReplyInputPortValueRetrieved);
					  char *s = vuoReceiveAndCopyString(ZMQControl);
					  valueAsString = s;
					  free(s);
				  });
	return valueAsString;
}

/**
 * Sends a control request to the composition to get an output port's value.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 *
 * @param portIdentifier The compile-time identifier for the port (see @c VuoCompilerEventPort::getIdentifier()).
 * @return String representation of the port's value.
 *
 * @see VuoTypes for information about types and serialization.
 */
string VuoRunner::getOutputPortValue(string portIdentifier)
{
	__block string valueAsString;
	dispatch_sync(controlQueue, ^{
					  vuoMemoryBarrier();

					  zmq_msg_t messages[2];
					  vuoInitMessageWithBool(&messages[0], !isInCurrentProcess());
					  vuoInitMessageWithString(&messages[1], portIdentifier.c_str());
					  vuoControlRequestSend(VuoControlRequestOutputPortValueRetrieve, messages, 2);
					  vuoControlReplyReceive(VuoControlReplyOutputPortValueRetrieved);
					  char *s = vuoReceiveAndCopyString(ZMQControl);
					  valueAsString = s;
					  free(s);
				  });
	return valueAsString;
}

/**
 * Sends a control request to the composition to get an input port's summary.
 *
 * Assumes the composition has been started and has not been stopped.
 *
 * @param portIdentifier The compile-time identifier for the port (see @c VuoCompilerEventPort::getIdentifier()).
 * @return A brief description of the port's value.
 *
 * @see VuoTypes for information about types and serialization.
 */
string VuoRunner::getInputPortSummary(string portIdentifier)
{
	__block string summary;
	dispatch_sync(controlQueue, ^{
					  vuoMemoryBarrier();

					  zmq_msg_t messages[1];
					  vuoInitMessageWithString(&messages[0], portIdentifier.c_str());
					  vuoControlRequestSend(VuoControlRequestInputPortSummaryRetrieve, messages, 1);
					  vuoControlReplyReceive(VuoControlReplyInputPortSummaryRetrieved);
					  char *s = vuoReceiveAndCopyString(ZMQControl);
					  summary = s;
					  free(s);
				  });
	return summary;
}

/**
 * Sends a control request to the composition to get an output port's summary.
 *
 * Assumes the composition has been started and has not been stopped.
 *
 * @param portIdentifier The compile-time identifier for the port (see @c VuoCompilerEventPort::getIdentifier()).
 * @return A brief description of the port's value.
 *
 * @see VuoTypes for information about types and serialization.
 */
string VuoRunner::getOutputPortSummary(string portIdentifier)
{
	__block string summary;
	dispatch_sync(controlQueue, ^{
					  vuoMemoryBarrier();

					  zmq_msg_t messages[1];
					  vuoInitMessageWithString(&messages[0], portIdentifier.c_str());
					  vuoControlRequestSend(VuoControlRequestOutputPortSummaryRetrieve, messages, 1);
					  vuoControlReplyReceive(VuoControlReplyOutputPortSummaryRetrieved);
					  char *s = vuoReceiveAndCopyString(ZMQControl);
					  string summary = s;
					  free(s);
				  });
	return summary;
}

/**
 * Sends a control request to the composition telling it to modify a published input port's value.
 * When this method returns, the input port value will have been modified.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 *
 * @see VuoTypes for information about types and serialization.
 */
void VuoRunner::setPublishedInputPortValue(VuoRunner::Port *port, string valueAsString)
{
	set<string> connectedPortIdentifiers = port->getConnectedPortIdentifiers();
	for (set<string>::iterator i = connectedPortIdentifiers.begin(); i != connectedPortIdentifiers.end(); ++i)
	{
		setInputPortValue((*i).c_str(), valueAsString);
	}
}

/**
 * Sends a control request to the composition telling it to generate an event on a published input port.
 * When this method returns, the event will have been enqueued.
 * This method clears the event flags used by @c waitForAnyPublishedOutputPortEvent().
 *
 * Assumes the composition has been started, is not paused, and has not been hibernated or stopped.
 */
void VuoRunner::generatePublishedInputPortEvent(VuoRunner::Port *port)
{
	dispatch_sync(controlQueue, ^{
					  vuoMemoryBarrier();

					  anyPublishedPortEventSignaled = false;

					  zmq_msg_t messages[1];
					  vuoInitMessageWithString(&messages[0], port->getName().c_str());
					  vuoControlRequestSend(VuoControlRequestPublishedInputPortGenerateEvent, messages, 1);
					  vuoControlReplyReceive(VuoControlReplyPublishedInputPortFiredEvent);
				  });
}

/**
 * Waits until the first event following a call to @c generatePublishedInputPortEvent() emerges from any published output port.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 *
 * @eg{
 *   runner.setPublishedInputPortValue(inputPort, inputValue);
 *   runner.setPublishedInputPortValue(anotherInputPort, anotherInputValue);
 *   runner.generatePublishedInputPortEvent(inputPort);
 *   runner.waitForAnyPublishedOutputPortEvent();
 *   result = runner.getPublishedOutputPortValue(outputPort);
 * }
 */
void VuoRunner::waitForAnyPublishedOutputPortEvent(void)
{
	saturating_semaphore_wait(anyPublishedPortEventSemaphore, &anyPublishedPortEventSignaled);
}

/**
 * Sends a control request to the composition to get a published output port's value.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 *
 * @see VuoTypes for information about types and serialization.
 */
string VuoRunner::getPublishedOutputPortValue(VuoRunner::Port *port)
{
	set<string> connectedPortIdentifiers = port->getConnectedPortIdentifiers();
	return getOutputPortValue( connectedPortIdentifiers.begin()->c_str() );
}

/**
 * Returns the list of published input or output ports for the running composition.
 *
 * The first time this method is called for inputs and for outputs, it sends a control request to the composition.
 * On subsequent calls, it returns the list of ports cached from the first call.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 *
 * If a control request is sent, this function must be synchronized with other control requests
 * (i.e. called from @c start(bool) or @c controlQueue).
 */
vector<VuoRunner::Port *> VuoRunner::getCachedPublishedPorts(bool input)
{
	// Caching not only provides faster access (without zmq messages),
	// but also ensures that the VuoRunner::Port pointers passed to
	// VuoRunnerDelegate::receivedTelemetryPublishedOutputPortUpdated are consistent.

	if (input)
	{
		if (! arePublishedInputPortsCached)
		{
			publishedInputPorts = refreshPublishedPorts(true);
			arePublishedInputPortsCached = true;
		}
		return publishedInputPorts;
	}
	else
	{
		if (! arePublishedOutputPortsCached)
		{
			publishedOutputPorts = refreshPublishedPorts(false);

			for (vector<VuoRunner::Port *>::iterator i = publishedOutputPorts.begin(); i != publishedOutputPorts.end(); ++i)
			{
				VuoRunner::Port *port = *i;
				set<string> connectedIdentifiers = port->getConnectedPortIdentifiers();
				publishedOutputPortForIdentifier[ *connectedIdentifiers.begin() ] = port;
			}

			arePublishedOutputPortsCached = true;
		}
		return publishedOutputPorts;
	}
}

/**
 * Sends a control request to the composition to get a list of published input or output ports.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 *
 * This function must be synchronized with other control requests
 * (i.e. called from @c start(bool) or @c controlQueue).
 */
vector<VuoRunner::Port *> VuoRunner::refreshPublishedPorts(bool input)
{
	vuoMemoryBarrier();

	enum VuoControlRequest requests[3];
	enum VuoControlReply replies[3];
	if (input)
	{
		requests[0] = VuoControlRequestPublishedInputPortNamesRetrieve;
		requests[1] = VuoControlRequestPublishedInputPortTypesRetrieve;
		requests[2] = VuoControlRequestPublishedInputPortConnectedIdentifiersRetrieve;
		replies[0] = VuoControlReplyPublishedInputPortNamesRetrieved;
		replies[1] = VuoControlReplyPublishedInputPortTypesRetrieved;
		replies[2] = VuoControlReplyPublishedInputPortConnectedIdentifiersRetrieved;
	}
	else
	{
		requests[0] = VuoControlRequestPublishedOutputPortNamesRetrieve;
		requests[1] = VuoControlRequestPublishedOutputPortTypesRetrieve;
		requests[2] = VuoControlRequestPublishedOutputPortConnectedIdentifiersRetrieve;
		replies[0] = VuoControlReplyPublishedOutputPortNamesRetrieved;
		replies[1] = VuoControlReplyPublishedOutputPortTypesRetrieved;
		replies[2] = VuoControlReplyPublishedOutputPortConnectedIdentifiersRetrieved;
	}

	vector<string> names;
	vector<string> types;

	for (int i = 0; i < 2; ++i)
	{
		vuoControlRequestSend(requests[i], NULL, 0);
		vuoControlReplyReceive(replies[i]);
		vector<string> messageStrings = receiveListOfStrings(ZMQControl);
		if (i == 0)
			names = messageStrings;
		else
			types = messageStrings;
	}

	vector<VuoRunner::Port *> ports;
	for (int i = 0; i < names.size() && i < types.size(); ++i)
	{
		zmq_msg_t messages[1];
		vuoInitMessageWithString(&messages[0], names[i].c_str());
		vuoControlRequestSend(requests[2], messages, 1);
		vuoControlReplyReceive(replies[2]);
		vector<string> messageStrings = receiveListOfStrings(ZMQControl);
		set<string> identifiers( messageStrings.begin(), messageStrings.end() );

		VuoRunner::Port *port = new Port(names[i], new VuoType(types[i]));
		port->setConnectedPortIdentifiers(identifiers);
		ports.push_back(port);
	}

	return ports;
}

/**
 * Returns the list of published input ports in the composition.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 */
vector<VuoRunner::Port *> VuoRunner::getPublishedInputPorts(void)
{
	return getCachedPublishedPorts(true);
}

/**
 * Returns the list of published output ports in the composition.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 */
vector<VuoRunner::Port *> VuoRunner::getPublishedOutputPorts(void)
{
	return getCachedPublishedPorts(false);
}

/**
 * Returns the published input port with the given name, or NULL if no such port exists.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 */
VuoRunner::Port * VuoRunner::getPublishedInputPortWithName(string name)
{
	vector<VuoRunner::Port *> inputPorts = getPublishedInputPorts();
	for (vector<VuoRunner::Port *>::iterator i = inputPorts.begin(); i != inputPorts.end(); ++i)
		if ((*i)->getName() == name)
			return *i;

	return NULL;
}

/**
 * Returns the published output port with the given name, or NULL if no such port exists.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 */
VuoRunner::Port * VuoRunner::getPublishedOutputPortWithName(string name)
{
	vector<VuoRunner::Port *> outputPorts = getPublishedOutputPorts();
	for (vector<VuoRunner::Port *>::iterator i = outputPorts.begin(); i != outputPorts.end(); ++i)
		if ((*i)->getName() == name)
			return *i;

	return NULL;
}

/**
 * Listens for telemetry data from the composition until the composition stops.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 */
void VuoRunner::listen(void)
{
	{
		ZMQTelemetry = zmq_socket(ZMQContext,ZMQ_SUB);
		if(zmq_connect(ZMQTelemetry,ZMQTelemetryURL.c_str()))
			fprintf(stderr, "runner VuoTelemetry: connect failed.\n");
	}

	{
		// subscribe to all types of telemetry
		char type = VuoTelemetryStats;
		zmq_setsockopt(ZMQTelemetry, ZMQ_SUBSCRIBE, &type, sizeof type);
		type = VuoTelemetryNodeExecutionStarted;
		zmq_setsockopt(ZMQTelemetry, ZMQ_SUBSCRIBE, &type, sizeof type);
		type = VuoTelemetryNodeExecutionFinished;
		zmq_setsockopt(ZMQTelemetry, ZMQ_SUBSCRIBE, &type, sizeof type);
		type = VuoTelemetryInputPortsUpdated;
		zmq_setsockopt(ZMQTelemetry, ZMQ_SUBSCRIBE, &type, sizeof type);
		type = VuoTelemetryOutputPortsUpdated;
		zmq_setsockopt(ZMQTelemetry, ZMQ_SUBSCRIBE, &type, sizeof type);
		type = VuoTelemetryEdgeHit;
		zmq_setsockopt(ZMQTelemetry, ZMQ_SUBSCRIBE, &type, sizeof type);
		type = VuoTelemetryError;
		zmq_setsockopt(ZMQTelemetry, ZMQ_SUBSCRIBE, &type, sizeof type);
	}

	// Wait for the first VuoTelemetryStats message to arrive, then confirm that we're able to receive messages
	// (if we don't wait for a message to arrive, important one-time-only messages such as VuoTelemetryOutputPortsUpdated might be lost).
	// As http://zguide.zeromq.org/page:all says, "the subscriber will always miss the first messages that the publisher sends".
	zmq_pollitem_t items[]=
	{
		{ZMQTelemetry,0,ZMQ_POLLIN,0},
	};
	int itemCount = 1;
	long timeout = -1;
	zmq_poll(items,itemCount,timeout);
	dispatch_semaphore_signal(beganListeningSemaphore);

	while(! listenCanceled)
	{
		long timeout = USEC_PER_SEC;  // wait up to 1 second (can't wait forever, since we need to check listenCanceled)
		zmq_poll(items,itemCount,timeout);
		if(items[0].revents & ZMQ_POLLIN)
		{
			// Receive telemetry type.
			char type = vuoReceiveInt(ZMQTelemetry);

			// Receive telemetry arguments and forward to VuoRunnerDelegate.
			switch (type)
			{
				case VuoTelemetryStats:
				{
					unsigned long utime = vuoReceiveUnsignedLong(ZMQTelemetry);
					unsigned long stime = vuoReceiveUnsignedLong(ZMQTelemetry);
					dispatch_sync(delegateQueue, ^{
									  if (delegate)
										  delegate->receivedTelemetryStats(utime, stime);
								  });
					break;
				}
				case VuoTelemetryNodeExecutionStarted:
				{
					char *nodeIdentifier = vuoReceiveAndCopyString(ZMQTelemetry);
					dispatch_sync(delegateQueue, ^{
									  if (delegate)
										  delegate->receivedTelemetryNodeExecutionStarted( string(nodeIdentifier) );
								  });
					free(nodeIdentifier);
					break;
				}
				case VuoTelemetryNodeExecutionFinished:
				{
					char *nodeIdentifier = vuoReceiveAndCopyString(ZMQTelemetry);
					dispatch_sync(delegateQueue, ^{
									  if (delegate)
										  delegate->receivedTelemetryNodeExecutionFinished( string(nodeIdentifier) );
								  });
					free(nodeIdentifier);
					break;
				}
				case VuoTelemetryEdgeHit:
				{
					char *edgeIdentifier = vuoReceiveAndCopyString(ZMQTelemetry);
					dispatch_sync(delegateQueue, ^{
									  if (delegate)
										  delegate->receivedTelemetryEdgeHit( string(edgeIdentifier) );
								  });
					free(edgeIdentifier);
					break;
				}
				case VuoTelemetryInputPortsUpdated:
				{
					while (hasMoreToReceive(ZMQTelemetry))
					{
						char *portIdentifier = vuoReceiveAndCopyString(ZMQTelemetry);
						if (hasMoreToReceive(ZMQTelemetry))
						{
							bool receivedEvent = vuoReceiveBool(ZMQTelemetry);
							if (hasMoreToReceive(ZMQTelemetry))
							{
								bool receivedData = vuoReceiveBool(ZMQTelemetry);
								if (hasMoreToReceive(ZMQTelemetry))
								{
									char *portDataSummary = vuoReceiveAndCopyString(ZMQTelemetry);
									dispatch_sync(delegateQueue, ^{
													  if (delegate)
														  delegate->receivedTelemetryInputPortUpdated(portIdentifier, receivedEvent, receivedData, portDataSummary);
												  });
									free(portDataSummary);
								}
							}
						}
						free(portIdentifier);
					}
					break;
				}
				case VuoTelemetryOutputPortsUpdated:
				{
					while (hasMoreToReceive(ZMQTelemetry))
					{
						char *portIdentifier = vuoReceiveAndCopyString(ZMQTelemetry);
						if (hasMoreToReceive(ZMQTelemetry))
						{
							bool sentData = vuoReceiveBool(ZMQTelemetry);
							if (hasMoreToReceive(ZMQTelemetry))
							{
								char *portDataSummary = vuoReceiveAndCopyString(ZMQTelemetry);
								dispatch_sync(delegateQueue, ^{
												  if (delegate)
													  delegate->receivedTelemetryOutputPortUpdated(portIdentifier, sentData, portDataSummary);
											  });

								map<string, VuoRunner::Port *>::iterator portIter = publishedOutputPortForIdentifier.find(portIdentifier);
								if (portIter != publishedOutputPortForIdentifier.end())
								{
									VuoRunner::Port *port = portIter->second;
									dispatch_sync(delegateQueue, ^{
													  if (delegate)
														  delegate->receivedTelemetryPublishedOutputPortUpdated(port, sentData, portDataSummary);
												  });

									saturating_semaphore_signal(anyPublishedPortEventSemaphore, &anyPublishedPortEventSignaled);
								}
								free(portDataSummary);
							}
						}
						free(portIdentifier);
					}
					break;
				}
				case VuoTelemetryError:
				{
					char *message = vuoReceiveAndCopyString(ZMQTelemetry);
					dispatch_sync(delegateQueue, ^{
									  if (delegate)
										  delegate->receivedTelemetryError( string(message) );
								  });
					free(message);
					break;
				}
			}
		}
	}

	zmq_close(ZMQTelemetry);
	ZMQTelemetry = NULL;

	dispatch_semaphore_signal(endedListeningSemaphore);
}

/**
 * Sends a control request to the composition via the control socket.
 *
 * Assumes the composition has been started and has not been hibernated or stopped.
 *
 * @param request The type of control request.
 * @param messages Arguments passed with the control request.
 * @param messageCount The number of arguments.
 */
void VuoRunner::vuoControlRequestSend(enum VuoControlRequest request, zmq_msg_t *messages, unsigned int messageCount)
{
	vuoSend("runner VuoControl",ZMQControl,request,messages,messageCount,false);
}

/**
 * Sends a control request to the composition loader via the composition loader's control socket.
 *
 * Assumes the composition has been started and has not been stopped.
 *
 * @param request The type of control request.
 * @param messages Arguments passed with the control request.
 * @param messageCount The number of arguments.
 */
void VuoRunner::vuoLoaderControlRequestSend(enum VuoLoaderControlRequest request, zmq_msg_t *messages, unsigned int messageCount)
{
	vuoSend("runner VuoLoaderControl",ZMQLoaderControl,request,messages,messageCount,false);
}

/**
 * Receives a control reply from the composition via the control socket, and checks that it
 * is the expected reply.
 *
 * Assumes the reply is the next message to be received on the control socket.
 *
 * @param expectedReply The expected reply value.
 */
void VuoRunner::vuoControlReplyReceive(enum VuoControlReply expectedReply)
{
	int reply = vuoReceiveInt(ZMQControl);
	if (reply != expectedReply)
		fprintf(stderr, "VuoControl message received unexpected reply (received %u, expected %u)\n", reply, expectedReply);
}

/**
 * Receives a control reply from the composition loader via the composition loader's control socket,
 * and checks that it is the expected reply.
 *
 * Assumes the reply is the next message to be received on the composition loader's control socket.
 *
 * @param expectedReply The expected reply value.
 */
void VuoRunner::vuoLoaderControlReplyReceive(enum VuoLoaderControlReply expectedReply)
{
	int reply = vuoReceiveInt(ZMQLoaderControl);
	if (reply != expectedReply)
		fprintf(stderr, "VuoLoaderControl message received unexpected reply (received %u, expected %u)\n", reply, expectedReply);
}

/**
 * Returns true if there are more messages to receive on the socket currently.
 */
bool VuoRunner::hasMoreToReceive(void *socket)
{
	int64_t more;
	size_t moreSize = sizeof more;
	zmq_getsockopt(socket, ZMQ_RCVMORE, &more, &moreSize);
	return more;
}

/**
 * Receives a list of 0 or more strings from the composition.
 *
 * @return The list of strings.
 */
vector<string> VuoRunner::receiveListOfStrings(void *socket)
{
	vector<string> messageStrings;
	while (hasMoreToReceive(socket))
	{
		char *s = vuoReceiveAndCopyString(socket);
		messageStrings.push_back(string(s));
		free(s);
	}
	return messageStrings;
}

/**
 * Signals the semaphore only if it's being waited on. Use with saturating_semaphore_wait().
 *
 * Copied from https://developer.apple.com/library/mac/#documentation/Darwin/Reference/ManPages/man3/dispatch_semaphore_create.3.html
 */
void VuoRunner::saturating_semaphore_signal(dispatch_semaphore_t dsema, bool *signaled)
{
	if (__sync_bool_compare_and_swap(signaled, false, true)) {
		dispatch_semaphore_signal(dsema);
	}
}

/**
 * Waits on the semaphore. Use with saturating_semaphore_signal().
 *
 * Copied from https://developer.apple.com/library/mac/#documentation/Darwin/Reference/ManPages/man3/dispatch_semaphore_create.3.html
 */
void VuoRunner::saturating_semaphore_wait(dispatch_semaphore_t dsema, bool *signaled)
{
	*signaled = false;
	dispatch_semaphore_wait(dsema, DISPATCH_TIME_FOREVER);
}

/**
 * Returns true if the composition either has not been started or has been started and stopped.
 */
bool VuoRunner::isStopped(void)
{
	return stopped;
}

/**
 * Returns true if the composition was constructed to run in the current process (even if the composition is not running right now).
 */
bool VuoRunner::isInCurrentProcess(void)
{
	return executablePath.empty();
}

/**
 * Returns true if the composition was constructed to run in a separate process with a composition loader
 * (even if the composition is not running right now).
 */
bool VuoRunner::isUsingCompositionLoader(void)
{
	return ! executablePath.empty() && ! dylibPath.empty();
}

/**
 * Sets the delegate that receives telemetry data. May be null.
 */
void VuoRunner::setDelegate(VuoRunnerDelegate *delegate)
{
	dispatch_sync(delegateQueue, ^{
					  this->delegate = delegate;
				  });
}

/**
 * Creates a dummy published port that is not yet connected to any port in a running composition.
 *
 * @param name The published port's name.
 * @param type The published port's data type, or null if the port is event-only.
 */
VuoRunner::Port::Port(string name, VuoType *type)
{
	this->name = name;
	this->type = type;
}

/**
 * Returns the published port's name.
 */
string VuoRunner::Port::getName(void)
{
	return name;
}

/**
 * Returns the published port's type, or null if the port is event-only.
 */
VuoType * VuoRunner::Port::getType(void)
{
	return type;
}

/**
 * Makes this published port an alias for the ports with the given identifiers.
 */
void VuoRunner::Port::setConnectedPortIdentifiers(set<string> connectedPortIdentifiers)
{
	this->connectedPortIdentifiers = connectedPortIdentifiers;
}

/**
 * Returns the identifiers of the ports for which this published port is an alias.
 */
set<string> VuoRunner::Port::getConnectedPortIdentifiers(void)
{
	return connectedPortIdentifiers;
}

#pragma mark VuoRunnerDelegate

VuoRunnerDelegate::~VuoRunnerDelegate() { }  // Fixes "undefined symbols" error
