﻿/**
 * @file
 * VuoCompiler interface.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This interface description may be modified and distributed under the terms of the GNU Lesser General Public License (LGPL) version 2 or later.
 * For more information, see http://vuo.org/license.
 */

#ifndef VUOCOMPILER_H
#define VUOCOMPILER_H

class VuoCompilerNodeClass;
#include "VuoCompilerInputEventPortClass.hh"
#include "VuoCompilerOutputEventPortClass.hh"
#include "VuoCompilerTriggerPortClass.hh"
#include "VuoCompilerNode.hh"
#include "VuoCompilerBitcodeGenerator.hh"
#include "VuoCompilerType.hh"
#include "VuoRunner.hh"


/**
 * A compiler/linker for node classes and compositions.
 *
 * Once you've compiled and linked a composition, you can use VuoRunner to execute and communicate with it.
 */
class VuoCompiler
{
private:
	map<string, VuoCompilerNodeClass *> nodeClasses;
	map<string, VuoCompilerType *> types;
	map<string, bool> isNodeClassAndTypeSearchPathLoaded;
	vector<string> nodeClassAndTypeSearchPaths;
	vector<string> headerSearchPaths;
	vector<string> librarySearchPaths;
	vector<string> preferredLibrarySearchPaths;
	vector<string> frameworkSearchPaths;
	vector<string> preferredFrameworkSearchPaths;
	llvm::sys::Path clangPath;
	string telemetry;
	string target;
	bool isVerbose;

	void loadNodeClassesAndTypesIfNeeded(void);
	void loadNodeClassesAndTypes(string path);
	void reifyPortTypes(void);
	void linkCompositionToCreateExecutableOrDynamicLibrary(string compiledCompositionPath, string linkedCompositionPath, bool isDylib);
	set<string> getDependenciesForComposition(const string &compiledCompositionPath);
	void getDependenciesRecursively(const string &dependency, set<string> &dependencies);
	void getLinkerInputs(const set<string> &dependencies, set<Module *> &modules, set<string> &libraries, set<string> &frameworks);
	void link(string outputPath, const set<Module *> &modules, const set<string> &libraries, const set<string> &frameworks, bool isDylib);
	Module * readModuleFromC(string inputPath);
	static Module * readModuleFromBitcode(string inputPath);
	static bool writeModuleToBitcode(Module *module, string outputPath);
	void setTargetForModule(Module *module, string target = "");
	static vector<string> getCoreVuoDependencies(void);
	string getRuntimeMainDependency(void);
	string getRuntimeDependency(void);
	static string getModuleNameForPath(string path);
	llvm::sys::Path getClangPath(void);
	string getCompositionStubPath(void);

	friend class TestVuoCompilerBitcodeGenerator;

protected:
	void addModuleSearchPath(string path);

	friend class TestVuoCompiler;		///< TestVuoCompiler needs to add a search path for its own private nodes/types.
	friend class TestCompositionExecution;	///< TestCompositionExecution needs to add a search path for its own private nodes.

public:
	VuoCompiler(void);
	static llvm::sys::Path getVuoFrameworkPath();
	static string getUserModulesPath();
	static string getSystemModulesPath();
	void addNodeClass(VuoCompilerNodeClass *nodeClass);
	void compileComposition(VuoCompilerBitcodeGenerator *generator, string outputPath);
	void compileComposition(string inputPath, string outputPath);
	void compileCompositionString(const string &compositionString, string outputPath);
	void compileNodeClass(string inputPath, string outputPath);
	void linkCompositionToCreateExecutable(string inputPath, string outputPath);
	void linkCompositionToCreateDynamicLibrary(string inputPath, string outputPath);
	void linkCompositionToCreateDynamicLibraries(string compiledCompositionPath, string linkedCompositionPath, string &newLinkedResourcePath, vector<string> &alreadyLinkedResourcePaths, set<string> &alreadyLinkedResources);
	VuoCompilerNodeClass * getNodeClass(const string &id);
	map<string, VuoCompilerNodeClass *> getNodeClasses(void);
	VuoCompilerType * getType(const string &id);
	void listNodeClasses(const string &format = "");
	void addHeaderSearchPath(const string &path);
	void addLibrarySearchPath(const string &path);
	void addPreferredLibrarySearchPath(const string &path);
	void clearPreferredLibrarySearchPaths();
	void addFrameworkSearchPath(const string &path);
	void addPreferredFrameworkSearchPath(const string &path);
	void setTelemetry(const string &telemetry);
	void setTarget(const string &target);
	void setVerbose(bool isVerbose);
	void setClangPath(const string &clangPath);
	string getCompositionLoaderPath(void);
	void print(void);

	static VuoRunner * newSeparateProcessRunnerFromCompositionFile(string compositionFilePath);
	static VuoRunner * newSeparateProcessRunnerFromCompositionString(string composition);
	static VuoRunner * newCurrentProcessRunnerFromCompositionFile(string compositionFilePath);
	static VuoRunner * newCurrentProcessRunnerFromCompositionString(string composition);
};


#endif
