/**
 * @file
 * VuoCompilerModule implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code is proprietary, and may not be redistributed.
 * For more information, see http://vuo.org/license.
 */

#include "VuoCompilerModule.hh"
#include "VuoCompilerNodeClass.hh"
#include "VuoCompilerType.hh"
#include "VuoStringUtilities.hh"

/**
 * Creates a VuoCompilerModule associated with the given LLVM module,
 * as a pseudo compiler detail class of the given VuoModule.
 */
VuoCompilerModule::VuoCompilerModule(VuoModule *base, Module *module)
{
	this->base = base;
	this->module = module;
}

/**
 * Destructor.
 */
VuoCompilerModule::~VuoCompilerModule()
{
	/// @todo
}

/**
 * Instantiates a VuoCompilerModule (or child class) corresponding to the type of
 * VuoCompilerModule defined in the LLVM module. If no type of VuoCompilerModule is defined,
 * returns NULL.
 */
VuoCompilerModule * VuoCompilerModule::newModule(string moduleKey, Module *module)
{
	if (isModule(module))
	{
		VuoNodeClass *nodeClass = VuoCompilerNodeClass::newNodeClass(moduleKey, module);
		if (nodeClass)
			return nodeClass->getCompiler();

		VuoCompilerType *type = VuoCompilerType::newType(moduleKey, module);
		return type;
	}
	return NULL;
}

/**
 * Returns true if the LLVM module contains certain global variables, indicating that it is
 * intended to be a VuoCompilerModule definition.
 */
bool VuoCompilerModule::isModule(Module *module)
{
	return (module->getNamedValue("moduleName") != NULL);
}

/**
 * Renames globals in the LLVM module that might conflict with other VuoCompilerModules,
 * and parses those globals.
 */
void VuoCompilerModule::parse(void)
{
	parser = new VuoCompilerBitcodeParser(module);
	renameGlobalVarsAndFuncs();
	parseMetadata();
}

/**
 * Parse the metadata of this VuoCompilerModule (name, description, ...) from the LLVM module.
 */
void VuoCompilerModule::parseMetadata(void)
{
	base->setDefaultDisplayName(parser->getGlobalString(	nameForGlobal("moduleName")));
	base->setDescription(parser->getGlobalString(			nameForGlobal("moduleDescription")));
	base->setVersion(parser->getGlobalUInt(					nameForGlobal("moduleVersion")));
	base->setKeywords(parser->getStringsFromGlobalArray(	nameForGlobal("moduleKeywords")));
	this->dependencies = parser->getStringsFromGlobalArray(	nameForGlobal("moduleDependencies"));
}

/**
 * Returns a list of global variables and functions that may be defined in multiple modules,
 * and thus need to be renamed.
 */
set<string> VuoCompilerModule::globalsToRename(void)
{
	set<string> globals;
	globals.insert("moduleName");
	globals.insert("moduleDescription");
	globals.insert("moduleVersion");
	globals.insert("moduleKeywords");
	globals.insert("moduleDependencies");
	return globals;
}

/**
 * Returns the mangled name for a function or global variable.
 */
string VuoCompilerModule::nameForGlobal(string genericGlobalVarOrFuncName)
{
	string moduleKey = base->getModuleKey();
	string globalVarOrFuncName = VuoStringUtilities::transcodeToIdentifier(moduleKey) + "__" + genericGlobalVarOrFuncName;
	return globalVarOrFuncName;
}

/**
 * Renames the global variables and functions within the LLVM module so that they are
 * unique to this VuoCompilerModule.
 */
void VuoCompilerModule::renameGlobalVarsAndFuncs(void)
{
	set<string> globalsAndFuncsToRename = globalsToRename();

	Module::GlobalListType& globals = module->getGlobalList();

	// Iterate through global variables
	for (Module::GlobalListType::iterator i = globals.begin(), e = globals.end(); i != e; ++i) {
		string curGlobalVarName = i->getName();
		// If current global variable is in the list of those to rename, do so
		if (globalsAndFuncsToRename.find(curGlobalVarName) != globalsAndFuncsToRename.end()) {
			string newGlobalVarName = nameForGlobal(curGlobalVarName);
			i->setName(newGlobalVarName);
		}

	}

	Module::FunctionListType& functions = module->getFunctionList();

	// Iterate through functions
	for (Module::FunctionListType::iterator i = functions.begin(), e = functions.end(); i != e; ++i) {
		string curFuncName = i->getName();
		// If current function is in the list of those to rename, do so
		if (globalsAndFuncsToRename.find(curFuncName) != globalsAndFuncsToRename.end()) {
			string newFuncName = nameForGlobal(curFuncName);
			i->setName(newFuncName);
		}

	}
}

/**
 * Copies the function's header into the LLVM module, if it's not already there.
 * Use this when an LLVM module needs to call a function defined in another LLVM module.
 */
Function * VuoCompilerModule::declareFunctionInModule(Module *module, Function *functionSrc)
{
	Function *functionDst = module->getFunction(functionSrc->getName());
	if (! functionDst)
	{
		functionDst = Function::Create(functionSrc->getFunctionType(),
									   GlobalValue::ExternalLinkage,
									   functionSrc->getName(),
									   module);
		functionDst->setAttributes( functionSrc->getAttributes() );
	}
	return functionDst;
}

/**
 * Returns a list of this VuoCompilerModule's dependencies.
 */
vector<string> VuoCompilerModule::getDependencies(void)
{
	return dependencies;
}

/**
 * Returns this VuoCompilerModule's LLVM module, which other LLVM modules can link to.
 */
Module * VuoCompilerModule::getModule(void)
{
	return module;
}
