/**
 * @file
 * VuoCompilerType implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code is proprietary, and may not be redistributed.
 * For more information, see http://vuo.org/license.
 */

#include "VuoCompilerCodeGenUtilities.hh"
#include "VuoCompilerType.hh"

/**
 * Creates a type from an LLVM module, and creates its corresponding base @c VuoType.
 */
VuoCompilerType::VuoCompilerType(string typeName, Module *module)
	: VuoBaseDetail<VuoType>("VuoCompilerType", new VuoType(typeName)),
	  VuoCompilerModule(getBase(), module)
{
	getBase()->setCompiler(this);

	valueFromStringFunction = NULL;
	stringFromValueFunction = NULL;
	summaryFromValueFunction = NULL;

	parse();
}

/**
 * Creates a type from the type definition in the module. If the module does not contain
 * a type definition, returns NULL.
 */
VuoCompilerType * VuoCompilerType::newType(string typeName, Module *module)
{
	if (! isType(typeName, module))
		return NULL;

	return new VuoCompilerType(typeName, module);
}

/**
 * Returns true if the LLVM module contains certain global variables, indicating that it is
 * intended to be a type definition.
 */
bool VuoCompilerType::isType(string typeName, Module *module)
{
	return (module->getNamedValue(typeName + "_valueFromString") != NULL);
}

/**
 * Overrides the implementation in @c VuoCompilerModule.
 */
void VuoCompilerType::parse(void)
{
	VuoCompilerModule::parse();

	string typeName = getBase()->getModuleKey();
	valueFromStringFunction = parser->getFunction(typeName + "_valueFromString");
	stringFromValueFunction = parser->getFunction(typeName + "_stringFromValue");
	interprocessStringFromValueFunction = parser->getFunction(typeName + "_interprocessStringFromValue");
	summaryFromValueFunction = parser->getFunction(typeName + "_summaryFromValue");

	if (! valueFromStringFunction)
		fprintf(stderr, "Couldn't find %s_valueFromString() function\n", typeName.c_str());
	if (! stringFromValueFunction)
		fprintf(stderr, "Couldn't find %s_stringFromValue() function\n", typeName.c_str());
	if (! summaryFromValueFunction)
		fprintf(stderr, "Couldn't find %s_summaryFromValueFunction() function\n", typeName.c_str());

	llvmType = VuoCompilerCodeGenUtilities::getParameterTypeBeforeLowering(stringFromValueFunction, 0, module, typeName);
}

/**
 * Overrides the implementation in @c VuoCompilerModule.
 */
set<string> VuoCompilerType::globalsToRename(void)
{
	set<string> globals = VuoCompilerModule::globalsToRename();
	return globals;
}

/**
 * Generates a call to @c [Type]_valueFromString().
 *
 * @param module The destination LLVM module (i.e., generated code).
 * @param block The LLVM block to which to append the function call.
 * @param arg The argument to pass to @c [Type]_valueFromString().
 * @return The return value of @c [Type]_valueFromString().
 */
Value * VuoCompilerType::generateValueFromStringFunctionCall(Module *module, BasicBlock *block, Value *arg)
{
	Function *function = declareFunctionInModule(module, valueFromStringFunction);

	if (VuoCompilerCodeGenUtilities::isFunctionReturningStructViaParameter(valueFromStringFunction))
	{
		vector<Value *> functionArgs;
		functionArgs.push_back(arg);
		Value *returnVariable = VuoCompilerCodeGenUtilities::callFunctionWithStructReturn(function, functionArgs, block);

		// Fix return type to match llvmType.
		returnVariable = new BitCastInst(returnVariable, PointerType::get(llvmType, 0), "", block);

		return new LoadInst(returnVariable, "", false, block);
	}
	else
	{
		Value *returnValue = CallInst::Create(function, arg, "", block);
		return VuoCompilerCodeGenUtilities::generateTypeCast(module, block, returnValue, llvmType);
	}
}

/**
 * Generates a call to @c [Type]_stringFromValue().
 *
 * @param module The destination LLVM module (i.e., generated code).
 * @param block The LLVM block to which to append the function call.
 * @param arg The argument to pass to @c [Type]_stringFromValue().
 * @return The return value of @c [Type]_stringFromValue().
 */
Value * VuoCompilerType::generateStringFromValueFunctionCall(Module *module, BasicBlock *block, Value *arg)
{
	return generateSerializationFunctionCall(module, block, arg, stringFromValueFunction);
}

/**
 * Generates a call to @c [Type]_interprocessStringFromValue().
 *
 * @param module The destination LLVM module (i.e., generated code).
 * @param block The LLVM block to which to append the function call.
 * @param arg The argument to pass to @c [Type]_interprocessStringFromValue().
 * @return The return value of @c [Type]_interprocessStringFromValue(), or the return value of @c [Type]_stringFromValue() if the interprocess function doesn't exist.
 */
Value * VuoCompilerType::generateInterprocessStringFromValueFunctionCall(Module *module, BasicBlock *block, Value *arg)
{
	if (interprocessStringFromValueFunction)
		return generateSerializationFunctionCall(module, block, arg, interprocessStringFromValueFunction);
	else
		return generateStringFromValueFunctionCall(module, block, arg);
}

/**
 * Generates a call to @c [Type]_summaryFromValue().
 *
 * @param module The destination LLVM module (i.e., generated code).
 * @param block The LLVM block to which to append the function call.
 * @param arg The argument to pass to @c [Type]_summaryFromValue().
 * @return The return value of @c [Type]_summaryFromValue().
 */
Value * VuoCompilerType::generateSummaryFromValueFunctionCall(Module *module, BasicBlock *block, Value *arg)
{
	return generateSerializationFunctionCall(module, block, arg, summaryFromValueFunction);
}

/**
 * Generates a call to @c [Type]_stringFromValue() or @c [Type]_summaryFromValue().
 */
Value * VuoCompilerType::generateSerializationFunctionCall(Module *module, BasicBlock *block, Value *arg, Function *sourceFunction)
{
	Function *function = declareFunctionInModule(module, sourceFunction);

	Value *secondArgument = NULL;
	Value **secondArgumentIfNeeded = (function->getFunctionType()->getNumParams() == 2 ? &secondArgument : NULL);
	arg = VuoCompilerCodeGenUtilities::convertArgumentToParameterType(arg, function, 0, secondArgumentIfNeeded, module, block);

	vector<Value *> args;
	args.push_back(arg);
	if (secondArgument)
		args.push_back(secondArgument);
	return CallInst::Create(function, args, "", block);

}

/**
 * Returns the LLVM type for this Vuo type.
 */
Type * VuoCompilerType::getType(void)
{
	return llvmType;
}

/**
 * Returns the LLVM type for this Vuo type when it appears as a function parameter.
 *
 * This is needed, for example, for struct parameters with the "byval" attribute.
 */
Type * VuoCompilerType::getFunctionParameterType(void)
{
	return stringFromValueFunction->getFunctionType()->getParamType(0);
}

/**
 * Returns the LLVM attributes for this Vuo type when it appears as a function parameter.
 *
 * This is needed, for example, for struct parameters with the "byval" attribute.
 */
Attributes VuoCompilerType::getFunctionParameterAttributes(void)
{
	return stringFromValueFunction->getAttributes().getParamAttributes(1);
}
