@addtogroup WritingNodeClasses

To write a node class in C, you need to: 

  - <code>\#include "node.h"</code>
    - This file contains macros needed for writing node classes. 
  - Define global variables for the node class's metadata. 
    - The metadata includes information like name, version, and keywords. 
  - Implement either the @ref nodeEvent function (for stateless node classes) or the @ref nodeInstanceEvent, @ref nodeInstanceInit, and @ref nodeInstanceFini functions (for stateful node classes). 
    - The @ref nodeEvent or @ref nodeInstanceEvent function is called when an event reaches one or more of the node's input ports, or in other words, when the node @term{executes}. This function can receive data and events through input ports and send data and events through output ports. 
    - The @ref nodeInstanceInit function is called once when the composition starts running. It initializes the node's state. 
    - The @ref nodeInstanceFini function is called once when the composition stops running. It can be used for cleanup, if needed. 

Read on for details. 


## Getting started

The easiest way to begin a new node class is with the Qt Creator wizard. 

@todo: Qt Creator wizard for custom node classes - https://b33p.net/kosada/node/4015. Describe how to create a node class C file, how to build, how to install. 


## Stateless node classes

Stateless node classes are simpler, so let's start with them. A stateless node (for example, @vuoNode{Add} or @vuoNode{Read from Console}) can take in data and events through input ports. If it's an interface node, it can also see the world outside the composition (filesystem, network, devices). But when a stateless node executes, it can't look up information about previous executions. It has no way to store its state in memory. 

Let's walk through a simple example: a simplified version of the @vuoNode{Add} node class. We'll call it @vuoNode{Add Two Integers}. This node class will output the sum of two integers. 

The node class will be implemented in a file called `example.addTwoIntegers.c`. The file name determines the class's machine name (as opposed to its human-readable name, @vuoNode{Add Two Integers}). So the machine name is @vuoNodeClass{example.addTwoIntegers}. 

@code{.cc}
#include "node.h"
@endcode

Every node class needs this line. 

@code{.cc}
const char * moduleName = "Add Two Integers";
const char * moduleDescription = "Outputs the sum of the input integers.";
const unsigned int moduleVersion = 100;
const char * moduleKeywords[] = { "sum" };
@endcode

This is metadata to help Vuo users know how to use your node class. For more information, see node.h. 

@code{.cc}
const bool nodeIsInterface = false;
@endcode

This is metadata to indicate how your node class interacts with the outside world. For more information, see node.h. 

@code{.cc}
void nodeEvent(
	VuoInputData(VuoInteger, {"default":0}) in0, 
	VuoInputData(VuoInteger, {"default":0}) in1, 
	VuoOutputData(VuoInteger) sum
) {
	*sum = in0 + in1;
}
@endcode

This function is called each time the node executes. The input and output ports are defined here, as parameters of this function. An @vuoNode{Add Two Integers} node has two input ports called @vuoPort{in0} and @vuoPort{in1} and an output port called @vuoPort{sum}. The function adds the two integers supplied by the input ports and sends their sum through the output port. 

The second argument of the `VuoInputData` macro, `{"default":0}`, assigns the input port a default value of 0. In a running composition, the input port data keeps its default value until the port receives some other data through an incoming cable or unless the port has been assigned a constant value. 

In addition to the port's default value, the second argument of the `VuoInputData` macro can specify other details about the port. These details can be used by input editor plugins in the Vuo Editor when the user is editing the constant value of the port. For example, `VuoInputData(VuoInteger, {"default":0,"suggestedMin":0})` would cause the port's input editor to display a spin box widget whose buttons have a minimum value of 0 (but whose text box has no limit). Some input editors support `suggestedMin`, `suggestedMax`, and `suggestedStep`. `suggestedMin` and `suggestedMax` provide a hint about the typical range of values for this port (but don't actually limit the values that the port can receive). `suggestedStep` is the suggested step amount for spin box and slider widgets. Some input editors may support additional details about the port. The details must be [JSON-formatted](http://www.json.org/). 

For more information on input and output ports, see node.h.

For more information on @c VuoInteger and other built-in port types, see Built-in Types.


## Stateful node classes

A stateful node class (for example, @vuoNode{Count} or @vuoNode{Recur}) keeps track of its state, which is stored in @term{node instance data}. The node instance data can be of any data type. It's returned by @ref nodeInstanceInit and passed as an argument to @ref nodeInstanceEvent and @ref nodeInstanceFini. 

Let's walk through another example: a simplified version of the @vuoNode{Count} node class. We'll call it @vuoNode{Count by One}. A @vuoNode{Count by One} node will output a count, which will be incremented whenever the node executes. 

The node class will be implemented in a file called `example.countByOne.c`, so its machine name will be @vuoNodeClass{example.countByOne}. 

@code{.cc}
#include "node.h"

const char * moduleName = "Count By One";
const char * moduleDescription = "Counts up by one when the refresh port receives an event.";
const unsigned int moduleVersion = 100;
const char * moduleKeywords[] = { "count", "increment" };

const bool nodeIsInterface = false;
@endcode

This is the node class's metadata. 

@code{.cc}
VuoInteger nodeInstanceInit()
{
	return 0;
}
@endcode

This function is called once for each @vuoNode{Count by One} node when the composition starts running. Its return type determines the type of the node instance data (@c VuoInteger). It returns 0, so each @vuoNode{Count by One} node starts off with a count of 0. 

@code{.cc}
void nodeInstanceEvent(
	VuoInstanceData(VuoInteger) countState,
	VuoOutputData(VuoInteger) count
) {
	*countState = *countState + 1;
	*count = *countState;
}
@endcode

This function is called each time the node executes. The function updates the node instance data (@c countState) and sends the updated count to the output port (@vuoPort{count}). 

@code{.cc}
void nodeInstanceFini(
	VuoInstanceData(VuoInteger) countState,
} {

}
@endcode

This function is called once for each @vuoNode{Count by One} node when the composition is stopped. The @ref nodeInstanceFini function can clean up after @ref nodeInstanceInit and @ref nodeInstanceEvent if needed (for example, closing files and sockets). For the @vuoNode{Count by One} node class, like most node classes, no cleanup is needed. 

Some node classes need to store heap-allocated data in their node instance data. For information on memory management, see @todo Managing Memory. 


## Naming node classes and ports

When naming node classes and ports, you're encouraged to follow these conventions: 

  - Node class names should:
    - be verb phrases (@vuoNode{Subtract}, not @vuoNode{Difference})
    - use [title case (Chicago Manual of Style)](http://blog.winepresspublishing.com/2012/06/grammar-tip-how-capitalize-titles/) (@vuoNode{Count within Range})
  - Node class file names should:
    - use lower camel case (@vuoNodeClass{vuo.math.isLessThan.integer})
    - include a lower camel case data type name as the last term, if and only if there are multiple versions of the node specializing in different data types
  - Port names should:
    - be full words or phrases, not abbreviations (@vuoPort{increment}, not @vuoPort{incr})
    - use lower-camel-case (@vuoPort{wrapMode})
  - Trigger port names should: 
    - be past-tense verb phrases (@vuoPort{started})
