@addtogroup WritingTypes

Vuo provides some built-in data types that you can use when writing node classes. For example, the @vuoNode{Make RGB Color} node class uses the @vuoType{VuoReal} type for its input ports and the @vuoType{VuoColor} type for its output port. If Vuo's built-in types are not enough, you can define your own custom type. 

To write a type in C, you need to: 

  - Define global variables for the type's metadata. 
    - The metadata includes information like name, version, and dependencies. 
  - Implement the `[your type]_jsonFromValue` and `[your type]_valueFromJson` functions to convert a value to and from a [JSON-C](https://github.com/json-c/json-c/wiki) object. 
  - Implement the `[your type]_summaryFromValue` function to provide a brief textual description of a value.
  - Optionally, implement the `[your type]_interprocessStringFromValue` function to convert a value from a string, if the conversion differs depending on whether the string came from the current process or a separate process. 

Read on for details. 


## Getting started

The easiest way to begin a new type is with the Qt Creator wizard. 

@todo: Qt Creator wizard for custom types - https://b33p.net/kosada/node/4015. Describe how to create a type header and implementation file, how to build, how to install. 


## Example

One reason you might need to define your own type is to allow a port to have one of a small set of values. For example, the @vuoType{VuoGradientNoise} type has two possible values: `VuoGradientNoise_Perlin` and `VuoGradientNoise_Simplex`. The @vuoNode{Make Gradient Noise 1D} node class has an input port of type @vuoType{VuoGradientNoise}, which allows the user to choose between Perlin noise and simplex noise. 

Let's look at the @vuoType{VuoGradientNoise} type as an example. 

VuoGradientNoise.h declares the type and the required functions: 

@code{.c}
typedef enum {
	VuoGradientNoise_Perlin,
	VuoGradientNoise_Simplex
} VuoGradientNoise;

VuoGradientNoise VuoGradientNoise_valueFromJson(struct json_object * js);
struct json_object * VuoGradientNoise_jsonFromValue(const VuoGradientNoise value);
char * VuoGradientNoise_summaryFromValue(const VuoGradientNoise value);
@endcode

Notice that each function name is prefixed by the type name, `VuoGradientNoise`. 

VuoGradientNoise.c defines the functions and provides some additional information. Let's look at some of its code. 

@code{.c}
#include "type.h"
@endcode

Every type needs this line. Among other things, type.h includes the JSON-C library. 

@code{.cc}
#ifdef VUO_COMPILER
const char * moduleName = "Gradient Noise";
const char * moduleDescription = "A method for generating gradient noise.";
const unsigned int moduleVersion = 100;
const char * moduleKeywords[] = { "perlin", "simplex" };
const char * moduleDependencies[] = { "c" };
#endif
@endcode

This is metadata to help developers know how to use your type. For more information, see type.h. 

@code{.c}
VuoGradientNoise VuoGradientNoise_valueFromJson(json_object *js)
{
	const char *valueAsString = "";
	if (json_object_get_type(js) == json_type_string)
		valueAsString = json_object_get_string(js);

	VuoGradientNoise gn = VuoGradientNoise_Perlin;

	if (strcmp(valueAsString, "perlin") == 0) {
		gn = VuoGradientNoise_Perlin;
	} else if (strcmp(valueAsString, "simplex") == 0) {
		gn = VuoGradientNoise_Simplex;
	}

	return gn;
}
@endcode

This function is used when creating a @vuoType{VuoGradientNoise} value from its serialized (string) representation. Vuo converts the string to a `json_object`, then passes the `json_object` to this function. This function is first called when the composition starts running, to initialize the value in each @vuoType{VuoGradientNoise} port. While the composition is running, this function is called again each time the composition receives a message from the VuoRunner to change an input port's value (for example, when a Vuo Editor user changes an input port value). 

For types with heap-allocated data (such as @vuoType{VuoText}, which contains a C string), the return value should be registered by calling `vuoRegister`. 

@code{.c}
json_object * VuoGradientNoise_jsonFromValue(const VuoGradientNoise value)
{
	char * valueAsString = "";

	switch (value) {
		case VuoGradientNoise_Perlin:
			valueAsString = "perlin";
			break;
		case VuoGradientNoise_Simplex:
			valueAsString = "simplex";
			break;
	}

	return json_object_new_string(valueAsString);
}
@endcode

This function is used when serializing a @vuoType{VuoGradientNoise} value to a string. Vuo calls this function to convert a value to a `json_object`, then Vuo converts the returned `json_object` to a string. While the composition is running, this function is called each time the composition receives a message from the VuoRunner to retrieve an output port's value. 

@code{.c}
char * VuoGradientNoise_summaryFromValue(const VuoGradientNoise value)
{
	switch (value) {
		case VuoGradientNoise_Perlin:
		default:
			return strdup("Perlin");
		case VuoGradientNoise_Simplex:
			return strdup("Simplex");
	}
}
@endcode

This function is used when providing a brief description of the value to the VuoRunner. This function is called each time a node with a @vuoType{VuoGradientNoise} port executes. Be sure to keep the returned string brief, so that it can be sent efficiently many times per second to the VuoRunner. 

The returned string should be heap-allocated (for example, using `malloc` or `strdup`), since it will be freed by the caller. 


## Intraprocess and interprocess serialization

@todo https://b33p.net/kosada/node/5611
