/**
 * @file
 * VuoMIDI implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "VuoMIDI.h"

#include "RtMidi.h"
#include <dispatch/dispatch.h>


/**
 * Returns a list of the available MIDI input devices.
 */
VuoList_VuoMIDIDevice VuoMIDI_getInputDevices(void)
{
	RtMidiIn *midiin = new RtMidiIn();
	unsigned int portCount = midiin->getPortCount();
	VuoList_VuoMIDIDevice inputDevices = VuoListCreate_VuoMIDIDevice();
	for (unsigned int i = 0; i < portCount; ++i)
		VuoListAppendValue_VuoMIDIDevice(inputDevices, VuoMIDIDevice_make(i, VuoText_make(midiin->getPortName(i).c_str()), true));
	delete midiin;
	return inputDevices;
}

/**
 * Returns a list of the available MIDI output devices.
 */
VuoList_VuoMIDIDevice VuoMIDI_getOutputDevices(void)
{
	RtMidiOut *midiout = new RtMidiOut();
	unsigned int portCount = midiout->getPortCount();
	VuoList_VuoMIDIDevice outputDevices = VuoListCreate_VuoMIDIDevice();
	for (unsigned int i = 0; i < portCount; ++i)
		VuoListAppendValue_VuoMIDIDevice(outputDevices, VuoMIDIDevice_make(i, VuoText_make(midiout->getPortName(i).c_str()), false));
	delete midiout;
	return outputDevices;
}

/**
 * Creates a manager for sending messages to a MIDI device.
 */
VuoMIDIOut VuoMIDIOut_make(VuoMIDIDevice md)
{
	if (md.isInput)
	{
		/// @todo https://b33p.net/kosada/node/4724
		fprintf(stderr, "VuoMIDI: The specified MIDI device (%s) isn't an output device.\n", VuoMIDIDevice_summaryFromValue(md));
		return NULL;
	}

	const char *vuoMIDIID = "VuoMIDIOut_make";

	RtMidiOut *midiout = new RtMidiOut();
	try
	{
		if (md.id == -1 && strlen(md.name) == 0)
			// Open the first MIDI device
			midiout->openPort(0, vuoMIDIID);
		else if (md.id == -1)
		{
			// Open the first MIDI %s device whose name contains mn.name
			unsigned int portCount = midiout->getPortCount();
			for (unsigned int i = 0; i < portCount; ++i)
				if (midiout->getPortName(i).find(md.name) != std::string::npos)
				{
					midiout->openPort(i, vuoMIDIID);
					break;
				}
		}
		else
		{
			// Open the MIDI device specified by mn.id
			midiout->openPort(md.id, vuoMIDIID);
		}
	}
	catch (RtError &error)
	{
		/// @todo https://b33p.net/kosada/node/4724
		fprintf(stderr, "VuoMIDI: Failed to open the specified MIDI device (%s) :: %s.\n", VuoMIDIDevice_summaryFromValue(md), error.what());
		delete midiout;
		return NULL;
	}

	return (VuoMIDIOut)midiout;
}

/**
 * Outputs the specified @c note event through the specified MIDI output device @c mo.
 */
void VuoMIDIOut_sendNote(VuoMIDIOut mo, VuoMIDINote note)
{
	if (!mo)
		return;

	RtMidiOut *midiout = (RtMidiOut *)mo;
	std::vector<unsigned char> message;
	message.push_back(0x80 + (note.isNoteOn ? 0x10 : 0) + MIN(note.channel-1,15));
	message.push_back(MIN(note.noteNumber,127));
	message.push_back(MIN(note.velocity,127));
	midiout->sendMessage(&message);
}

/**
 * Outputs the specified @c controller event through the specified MIDI output device @c mo.
 */
void VuoMIDIOut_sendController(VuoMIDIOut mo, VuoMIDIController controller)
{
	if (!mo)
		return;

	RtMidiOut *midiout = (RtMidiOut *)mo;
	std::vector<unsigned char> message;
	message.push_back(0xB0 + MIN(controller.channel-1,15));
	message.push_back(MIN(controller.controllerNumber,127));
	message.push_back(MIN(controller.value,127));
	midiout->sendMessage(&message);
}

/**
 * Destroys a MIDI output device manager.
 */
void VuoMIDIOut_destroy(VuoMIDIOut mo)
{
	if (!mo)
		return;

	RtMidiOut *midiout = (RtMidiOut *)mo;
	delete midiout;
}

/**
 * Private data for a VuoMIDIIn instance.
 */
struct VuoMIDIIn_internal
{
	RtMidiIn *midiin;	///< RtMidi's device pointer.

	dispatch_queue_t callbackQueue;	///< Serializes access to the following callback functions.
	void (*receivedNote)(VuoMIDINote);	///< This node instance's trigger function.
	void (*receivedController)(VuoMIDIController);	///< This node instance's trigger function.
};

/**
 * RtMidi calls this function ("on a separate high-priority thread owned by CoreMIDI") every time it receives a MIDI event.
 */
void VuoMIDIIn_receivedEvent(double timeStamp, std::vector< unsigned char > *message, void *userData)
{
	struct VuoMIDIIn_internal *mii = (struct VuoMIDIIn_internal *)userData;
	unsigned char channel = ((*message)[0] & 0x0f) + 1;
	if (((*message)[0] & 0xf0) == 0x80) // Note Off
	{
		VuoMIDINote mn = VuoMIDINote_make(channel, false, (*message)[2], (*message)[1]);
		dispatch_async(mii->callbackQueue, ^{
						   if (mii->receivedNote)
								mii->receivedNote(mn);
					   });
	}
	else if (((*message)[0] & 0xf0) == 0x90) // Note On
	{
		VuoMIDINote mn = VuoMIDINote_make(channel, true, (*message)[2], (*message)[1]);
		dispatch_async(mii->callbackQueue, ^{
						   if (mii->receivedNote)
								mii->receivedNote(mn);
					   });
	}
	else if (((*message)[0] & 0xf0) == 0xb0) // Control Change
	{
		VuoMIDIController mc = VuoMIDIController_make(channel, (*message)[1], (*message)[2]);
		dispatch_async(mii->callbackQueue, ^{
						   if (mii->receivedController)
								mii->receivedController(mc);
					   });
	}
	else
	{
		const char *hex = "0123456789abcdef";
		std::string messageHex;
		for (size_t i = 0; i < message->size(); ++i)
		{
			messageHex.push_back(hex[(*message)[i] >> 4]);
			messageHex.push_back(hex[(*message)[i] & 0x0f]);
		}
		fprintf(stderr, "VuoMIDI: Received unknown message: 0x%s\n", messageHex.c_str());
	}
}

/**
 * Creates a manager for receiving messages from a MIDI device.
 */
VuoMIDIIn VuoMIDIIn_make(VuoMIDIDevice md)
{
	if (!md.isInput)
	{
		/// @todo https://b33p.net/kosada/node/4724
		fprintf(stderr, "VuoMIDI: The specified MIDI device (%s) isn't an input device.\n", VuoMIDIDevice_summaryFromValue(md));
		return NULL;
	}

	const char *vuoMIDIID = "VuoMIDIIn_make";

	struct VuoMIDIIn_internal *mii;
	RtMidiIn *midiin = new RtMidiIn();
	try
	{
		if (md.id == -1 && strlen(md.name) == 0)
			// Open the first MIDI device
			midiin->openPort(0, vuoMIDIID);
		else if (md.id == -1)
		{
			// Open the first MIDI %s device whose name contains mn.name
			unsigned int portCount = midiin->getPortCount();
			for (unsigned int i = 0; i < portCount; ++i)
				if (midiin->getPortName(i).find(md.name) != std::string::npos)
				{
					midiin->openPort(i, vuoMIDIID);
					break;
				}
		}
		else
		{
			// Open the MIDI device specified by mn.id
			midiin->openPort(md.id, vuoMIDIID);
		}

		mii = (struct VuoMIDIIn_internal *)calloc(1, sizeof(struct VuoMIDIIn_internal));
		mii->midiin = midiin;
		mii->callbackQueue = dispatch_queue_create("vuo.midi.receive", NULL);
		// Trigger functions (receiveNote, ...) are NULLed by calloc.

		midiin->setCallback(&VuoMIDIIn_receivedEvent, mii);

		// Ignore SysEx, timing, and active sensing for now.
		midiin->ignoreTypes(true,true,true);
	}
	catch (RtError &error)
	{
		/// @todo https://b33p.net/kosada/node/4724
		fprintf(stderr, "VuoMIDI: Failed to open the specified MIDI device (%s) :: %s.\n", VuoMIDIDevice_summaryFromValue(md), error.what());
		delete midiin;
		return NULL;
	}

	return (VuoMIDIIn)mii;
}

/**
 * Sets up the MIDI input device to call the trigger functions when it receives an event.
 *
 * Can be called from any thread.
 */
void VuoMIDIIn_enableTriggers
(
		VuoMIDIIn mi,
		VuoOutputTrigger(receivedNote, VuoMIDINote),
		VuoOutputTrigger(receivedController, VuoMIDIController)
)
{
	if (!mi)
		return;
	struct VuoMIDIIn_internal *mii = (struct VuoMIDIIn_internal *)mi;
	dispatch_async(mii->callbackQueue, ^{
					   mii->receivedNote = receivedNote;
					   mii->receivedController = receivedController;
				   });
}

/**
 * Stops the MIDI input device from calling trigger functions when it receives an event.
 *
 * Can be called from any thread.
 */
void VuoMIDIIn_disableTriggers(VuoMIDIIn mi)
{
	if (!mi)
		return;
	struct VuoMIDIIn_internal *mii = (struct VuoMIDIIn_internal *)mi;
	dispatch_sync(mii->callbackQueue, ^{
					   mii->receivedNote = NULL;
					   mii->receivedController = NULL;
				   });
}

/**
 * Destroys a MIDI input device manager.
 */
void VuoMIDIIn_destroy(VuoMIDIIn mi)
{
	if (!mi)
		return;
	struct VuoMIDIIn_internal *mii = (struct VuoMIDIIn_internal *)mi;
	delete mii->midiin;
}
