/**
 * @file
 * VuoURL implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"

#include <string.h>

#include <curl/curl.h>

/// @todo when https://b33p.net/kosada/node/4018 is fixed, move dependencies from vuo.image.get.c and vuo.scene.get.c.
//const char * moduleDependencies[] = {
//};

/**
 * Initializes the cURL library.
 */
__attribute__((constructor)) static void VuoURL_init(void)
{
	curl_global_init(CURL_GLOBAL_DEFAULT);
}

/**
 * A memory buffer, filled by @c VuoURL_curlCallback().
 */
struct VuoURL_curlBuffer
{
	char *memory;
	size_t size;
};

/**
 * Reads the contents of the URL into a memory buffer.
 * A callback function for use by @c curl_easy_setopt().
 */
static size_t VuoURL_curlCallback(void *contents, size_t size, size_t nmemb, void *userp)
{
	size_t realsize = size * nmemb;
	struct VuoURL_curlBuffer *mem = (struct VuoURL_curlBuffer *)userp;

	mem->memory = (char *)realloc(mem->memory, mem->size + realsize + 1);
	if(mem->memory == NULL)
	{
		fprintf(stderr, "VuoURL_curlCallback() Error: realloc() returned NULL (out of memory?).\n");
		return 0;
	}

	memcpy(&(mem->memory[mem->size]), contents, realsize);
	mem->size += realsize;
	mem->memory[mem->size] = 0;

	return realsize;
}

/**
 * Receives the data at the specified @c url.
 *
 * @return true upon success, false upon failure.
 * @todo Better error handling per https://b33p.net/kosada/node/4724
 */
bool VuoURL_get(const char *url, void **data, unsigned int *dataLength)
{
	struct VuoURL_curlBuffer buffer = {NULL, 0};
	CURL *curl;
	CURLcode res;

	curl = curl_easy_init();
	if (!curl)
	{
		fprintf(stderr, "VuoURL_get() Error: cURL initialization failed.\n");
		return false;
	}

	curl_easy_setopt(curl, CURLOPT_URL, url);
	curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);

	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, VuoURL_curlCallback);
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&buffer);

	res = curl_easy_perform(curl);
	if(res != CURLE_OK)
	{
		fprintf(stderr, "VuoURL_get() Error: cURL request failed: %s\n", curl_easy_strerror(res));
		return false;
	}

	curl_easy_cleanup(curl);

	*data = buffer.memory;
	*dataLength = buffer.size;
	return true;
}
