/**
 * @file
 * VuoWindow implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "VuoWindow.h"
#include "VuoWindowApplication.hh"
#include "VuoWindowTextInternal.hh"
#include "VuoWindowOpenGLInternal.hh"

extern "C"
{
/// @todo uncomment when https://b33p.net/kosada/node/4018 is fixed, and add all other VuoWindow*.bc and moc_VuoWindow*.bc modules as dependencies.
//const char *moduleDependencies[] = { "VuoGlContext", "QtCore.framework", "QtGui.framework" };
}


#include <mach-o/dyld.h>
#include <QDir>

/**
 * Returns the path to Vuo.framework.
 *
 * Partial clone of VuoCompiler::getVuoFrameworkPath(),
 * which we can't use here because the @c VuoCompiler class
 * doesn't get linked with built compositions.
 */
QString VuoWindowGetVuoFrameworkPath(void)
{
	// First check whether Vuo.framework is in the list of loaded dynamic libraries.
	for(unsigned int i=0; i<_dyld_image_count(); ++i)
	{
		QString name(_dyld_get_image_name(i));
		QString frameworkName("/Vuo.framework/Versions/" VUO_VERSION_STRING "/");
		int index;
		if ( (index = name.lastIndexOf(frameworkName)) > 0 )
		{
			QString nameWithoutFramework = name.left(index);
			nameWithoutFramework += "/Vuo.framework";
			QDir path(nameWithoutFramework);
			if (path.isReadable())
				return path.canonicalPath();
		}
	}

	fprintf(stderr, "Failed to find Vuo.framework.\n");
	return "";
}


/**
 * Creates a QApplication instance (if one doesn't already exist).
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowApplicationInit(void)
{
	dispatch_sync(dispatch_get_main_queue(), ^{
					  if (! QApplication::instance())
					  {
						  // Tell Qt where to find its plugins.
						  QApplication::setLibraryPaths(QStringList(VuoWindowGetVuoFrameworkPath() + "/Frameworks/QtPlugins"));

						  int argc = 0;
						  char ** argv = NULL;
						  QApplication * a = new VuoWindowApplication(argc,argv);

						  a->setQuitOnLastWindowClosed(true);
					  }
				  });
}

/**
 * Creates and displays a window containing a text edit widget.
 *
 * Creates a new QApplication instance if one did not already exist.
 *
 * Can be called from any thread but the main thread.
 */
VuoWindowText VuoWindowTextCreate(void)
{
	VuoWindowApplicationInit();

	__block VuoWindowTextInternal *window = NULL;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  window = new VuoWindowTextInternal();
					  window->show();
				  });
	return (VuoWindowText *)window;
}

/**
 * Sets up the window to call the trigger functions when its text is edited.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowTextEnableTriggers
(
		VuoWindowText w,
		VuoOutputTrigger(typedLine, VuoText),
		VuoOutputTrigger(typedWord, VuoText),
		VuoOutputTrigger(typedCharacter, VuoText)
)
{
	VuoWindowTextInternal *window = (VuoWindowTextInternal *)w;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  window->enableTriggers(typedLine, typedWord, typedCharacter);
				  });
}

/**
 * Stops the window from calling trigger functions when its text is edited.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowTextDisableTriggers(VuoWindowText w)
{
	VuoWindowTextInternal *window = (VuoWindowTextInternal *)w;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  window->disableTriggers();
				  });
}

/**
 * Creates and displays a window containing an OpenGL view.
 *
 * Creates a new QApplication instance if one did not already exist.
 *
 * Can be called from any thread but the main thread.
 */
VuoWindowOpenGL VuoWindowOpenGLCreate
(
		void (*initCallback)(void *),
		void (*resizeCallback)(void *, unsigned int, unsigned int),
		void (*drawCallback)(void *),
		void *context
)
{
	VuoWindowApplicationInit();

	__block VuoWindowOpenGLInternal *window = NULL;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  window = new VuoWindowOpenGLInternal(
							initCallback,
							resizeCallback,
							drawCallback,
							context
						);
					  window->show();
				  });
	return (VuoWindowOpenGL *)window;
}

/**
 * Sets up the window to call the trigger functions when events occur.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowOpenGLEnableTriggers
(
		VuoWindowOpenGL w,
		VuoOutputTrigger(frameRequested, VuoInteger), /// @todo change to struct...
		VuoOutputTrigger(mouseMoved, VuoPoint2d),
		VuoOutputTrigger(mouseLeftPressed, VuoPoint2d)
)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)w;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  window->enableTriggers(frameRequested, mouseMoved, mouseLeftPressed);
				  });
}

/**
 * Stops the window from calling trigger functions when events occur.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowOpenGLDisableTriggers(VuoWindowOpenGL w)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)w;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  window->disableTriggers();
				  });
}

/**
 * Appends text and a linebreak to the text edit widget in the window.
 *
 * Can be called from any thread.
 */
void VuoWindowTextAppendLine(VuoWindowText vw, const char *text)
{
	VuoWindowTextInternal *window = (VuoWindowTextInternal *)vw;
	dispatch_async(dispatch_get_main_queue(), ^{
					   window->appendLine(text);
				   });
}

/**
 * Schedules the specified window to be redrawn.
 *
 * Can be called from any thread.
 */
void VuoWindowOpenGLRedraw(VuoWindowOpenGL w)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)w;
	dispatch_async(dispatch_get_main_queue(), ^{
					   window->scheduleRedraw();
				   });
}

/**
 * Sets the window's current and preferred aspect ratio.
 *
 * If necessary, the window is resized to match the specified aspect ratio.
 * After calling this method, when the user resizes the window, its height is adjusted to match the specified aspect ratio.
 */
void VuoWindowOpenGLSetAspectRatio(VuoWindowOpenGL w, unsigned int pixelsWide, unsigned int pixelsHigh)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)w;
	dispatch_async(dispatch_get_main_queue(), ^{
					   window->setAspectRatio(pixelsWide, pixelsHigh);
				   });
}

/**
 * Closes and deallocates a @c VuoWindowText window.
 *
 * Destroys the current QApplication instance if it no longer contains any widgets.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowTextDestroy(VuoWindowText vw)
{
	VuoWindowTextInternal *window = (VuoWindowTextInternal *)vw;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  window->close();
					  delete window;
				  });
}

/**
 * Closes and deallocates a @c VuoWindowOpenGL window.
 *
 * Destroys the current QApplication instance if it no longer contains any widgets.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowOpenGLDestroy(VuoWindowOpenGL vw)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)vw;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  window->close();
					  delete window;
				  });
}
