/**
 * @file
 * VuoWindowOpenGLInternal implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "VuoWindowOpenGLInternal.hh"

#include <QCoreApplication>
#include <QGuiApplication>
#include <QResizeEvent>
#include <QScreen>

/**
 * Creates a window containing an OpenGL view.
 *
 * Must be called on the main thread.
 */
VuoWindowOpenGLInternal::VuoWindowOpenGLInternal
(
		void (*initCallback)(void *),
		void (*resizeCallback)(void *, unsigned int width, unsigned int height),
		void (*drawCallback)(void *),
		void *drawContext
)
	: QWindow()
{
	this->callbacksEnabled = false;
	this->initCallback = initCallback;
	this->resizeCallback = resizeCallback;
	this->drawCallback = drawCallback;
	this->drawContext = drawContext;

	this->frameRequested = NULL;
	this->mouseMoved = NULL;
	this->mouseLeftPressed = NULL;

	this->vuoGlContext = NULL;
	this->glContext = NULL;

	this->aspect = -1;
	this->userResizedWindow = false;

	setSurfaceType(QWindow::OpenGLSurface);

	setTitle("Vuo Scenegraph");

	this->programmaticallyResizingWindow = true;
	resize(1024,768);
	this->programmaticallyResizingWindow = false;

	// We can't activate the OpenGL context until the first @c exposeEvent(),
	// so we initialize in @c render().
}

/**
 * Throw this window's OpenGL context back into the pool.
 */
VuoWindowOpenGLInternal::~VuoWindowOpenGLInternal()
{
	if (glContext)
		delete glContext;

	// Deleting the QOpenGLContext destroys the underlying CGLContext, so we should _not_ call VuoGlContext_disuseSpecific() here.
}

/**
 * Creates and initializes the GL Context (if it doesn't already exist), and activates it on the current thread.
 */
void VuoWindowOpenGLInternal::ensureGLContext(void)
{
	if (!glContext)
	{
		vuoGlContext = VuoGlContext_use();
		glContext = new QOpenGLContext();
		if (!glContext->createWithNativeContext(vuoGlContext))
			fprintf(stderr,"VuoWindowOpenGLInternal::ensureGLContext() Failed to create GL context.\n");

		this->initCallback(this->drawContext);
	}

	if (!glContext->makeCurrent(this))
		fprintf(stderr,"VuoWindowOpenGLInternal::ensureGLContext() Failed to make GL context current.\n");
}

/**
 * Renders the OpenGL Window.
 */
void VuoWindowOpenGLInternal::render(void)
{
	if (! callbacksEnabled)
		return;

	ensureGLContext();

	this->drawCallback(this->drawContext);

	glContext->swapBuffers(this);
	glContext->doneCurrent();
}

/**
 * Override handling of certain events.
 */
bool VuoWindowOpenGLInternal::event(QEvent *event)
{
	switch (event->type())
	{
		case QEvent::UpdateRequest:
			render();
			return true;
		case QEvent::Close:
			// Prevent the window from closing.
			return true;
		default:
			return QWindow::event(event);
	}
}

/**
 * Render when the window is exposed.
 */
void VuoWindowOpenGLInternal::exposeEvent(QExposeEvent *event)
{
//	if (isExposed())
//		render();
}

/**
 * Returns true if the window's actual aspect ratio differs by more than 1 pixel from the ideal aspect ratio.
 */
bool VuoWindowOpenGLInternal::actualAspectDiffersSignificantlyFromIdeal(void)
{
	int actualHeight = height();
	int idealHeight = width()/aspect;
	return abs(actualHeight - idealHeight) > 1;
}

/**
 * Render when the window is resized.
 */
void VuoWindowOpenGLInternal::resizeEvent(QResizeEvent *event)
{
	if (! callbacksEnabled)
		return;

	if (aspect > 0 && actualAspectDiffersSignificantlyFromIdeal())
		resize(width(), width()/aspect);

	ensureGLContext();

	// Set the viewport to window dimensions
	glViewport(0, 0, width()*devicePixelRatio(), height()*devicePixelRatio());

	// Let the caller recalculate the projection matrix
	this->resizeCallback(this->drawContext, width()*devicePixelRatio(), height()*devicePixelRatio());

	if (isExposed())
		render();

	if (event && !this->programmaticallyResizingWindow)
		this->userResizedWindow = true;
}

/**
 * Sets up the window to call trigger functions and other callbacks when events occur.
 */
void VuoWindowOpenGLInternal::enableTriggers
(
		VuoOutputTrigger(frameRequested, VuoInteger), /// @todo change to struct...
		VuoOutputTrigger(mouseMoved, VuoPoint2d),
		VuoOutputTrigger(mouseLeftPressed, VuoPoint2d)
)
{
	callbacksEnabled = true;
	resizeEvent(NULL);  // The initial resize event was ignored, so make one now. Without it, nothing gets rendered.

	this->frameRequested = (void (*)(VuoInteger)) frameRequested; /// @todo change to struct...
	this->mouseMoved = (void (*)(VuoPoint2d)) mouseMoved;
	this->mouseLeftPressed = (void (*)(VuoPoint2d)) mouseLeftPressed;
	/// @todo actually fire these triggers
}

/**
 * Stops the window from calling trigger functions and other callbacks when events occur.
 */
void VuoWindowOpenGLInternal::disableTriggers(void)
{
	callbacksEnabled = false;

	this->frameRequested = NULL;
	this->mouseMoved = NULL;
	this->mouseLeftPressed = NULL;
}

/**
 * Schedules the GL widget to be redrawn.
 */
void VuoWindowOpenGLInternal::scheduleRedraw(void)
{
	QCoreApplication::postEvent(this, new QEvent(QEvent::UpdateRequest));
}

/**
 * Sets the window's current and preferred aspect ratio.
 *
 * If necessary, the window is resized to match the specified aspect ratio.
 * After calling this method, when the user resizes the window, its height is adjusted to match the specified aspect ratio.
 */
void VuoWindowOpenGLInternal::setAspectRatio(unsigned int pixelsWide, unsigned int pixelsHigh)
{
	this->aspect = (float)pixelsWide/(float)pixelsHigh;
	if (actualAspectDiffersSignificantlyFromIdeal())
	{
		unsigned int desiredWidth;
		unsigned int desiredHeight;

		if (this->userResizedWindow)
		{
			// If the user manually resized the window, preserve its width.
			desiredWidth = width();
			desiredHeight = width()/aspect;
		}
		else
		{
			// Otherwise match the exact width and height provided by the calling function.
			desiredWidth = pixelsWide;
			desiredHeight = pixelsHigh;
		}

		// Set the desired size, unless it's larger than the window manager's allowed size.
		QSize availableSize = ((QGuiApplication *)QGuiApplication::instance())->primaryScreen()->availableSize();
		float margin = 0.99;
		this->programmaticallyResizingWindow = true;
		if (desiredWidth <= availableSize.width()*margin && desiredHeight <= availableSize.height()*margin)
			resize(desiredWidth, desiredHeight);
		else if (desiredWidth > availableSize.width()*margin)
			// Too wide, so scale it to almost the maximum horizontal screen space.
			resize(availableSize.width()*margin, availableSize.width()*margin/aspect);
		else
			// Too tall, so scale it to almost the maximum vertical screen space.
			resize(availableSize.height()*margin*aspect, availableSize.height()*margin);
		this->programmaticallyResizingWindow = false;
	}
}
