/**
 * @file
 * VuoWindowTextInternal implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "VuoWindowTextInternal.hh"

#include <dispatch/dispatch.h>

/**
 * Creates a window containing a text edit widget.
 *
 * Must be called on the main thread.
 */
VuoWindowTextInternal::VuoWindowTextInternal(void)
{
	setWindowTitle("Vuo Console");
	setAttribute(Qt::WA_DeleteOnClose);

	pte = new QPlainTextEdit(this);
	setCentralWidget(pte);
	setFont(QFont("Monaco"));

	// Prevent the user from making changes other than appending text.
	pte->setUndoRedoEnabled(false);
	connect(pte, SIGNAL(cursorPositionChanged()), this, SLOT(moveCursorToEnd()));

	this->typedLine = NULL;
	this->typedWord = NULL;
	this->typedCharacter = NULL;

	resize(400,600);
}

/**
 * Sets up the window to call the trigger functions when its text is edited.
 */
void VuoWindowTextInternal::enableTriggers
(
		VuoOutputTrigger(typedLine, VuoText),
		VuoOutputTrigger(typedWord, VuoText),
		VuoOutputTrigger(typedCharacter, VuoText)
)
{
	this->typedLine = (void (*)(VuoText)) typedLine;
	this->typedWord = (void (*)(VuoText)) typedWord;
	this->typedCharacter = (void (*)(VuoText)) typedCharacter;
	connect(pte, SIGNAL(textChanged()), this, SLOT(fireTextChangedEvents()));
}

/**
 * Stops the window from calling trigger functions when its text is edited.
 */
void VuoWindowTextInternal::disableTriggers(void)
{
	this->typedLine = NULL;
	this->typedWord = NULL;
	this->typedCharacter = NULL;
	disconnect(pte, SIGNAL(textChanged()), this, SLOT(fireTextChangedEvents()));
}

/**
 * Appends text and a linebreak to the text edit widget.
 *
 * Must be called on the main thread.
 */
void VuoWindowTextInternal::appendLine(const char *text)
{
	pte->insertPlainText(QString(text) + QString('\n'));
	pte->ensureCursorVisible();
	lastPlainTextSet = pte->toPlainText();
}

/**
 * Calls the event trigger callbacks when the user has typed a line, word, or character in the text edit widget.
 *
 * Does not call the callbacks when text has been added by VuoWindowTextInternal::appendLine().
 */
void VuoWindowTextInternal::fireTextChangedEvents(void)
{
	dispatch_async(dispatch_get_main_queue(), ^{
					   int lengthDiff = pte->toPlainText().length() - lastPlainTextSet.length();
					   if (lengthDiff > 0)
					   {
						   QString stringDiff = pte->toPlainText().right(lengthDiff);

						   for (int i = 0; i < stringDiff.length(); ++i)
						   {
							   // Fire an event for each character added in the most recent edit.
							   QString currChar = stringDiff.mid(i, 1);
							   VuoText currCharAsVuoText = VuoText_make(currChar.toUtf8().constData());
							   typedCharacter(currCharAsVuoText);

							   if (currChar.at(0).isSpace())
							   {
								   // Fire an event for each word added in the most recent edit.
								   if (lastPlainTextSet.length() >= 1 && ! lastPlainTextSet.right(1).at(0).isSpace())
								   {
									   int lastWordBreak;
									   for (lastWordBreak = lastPlainTextSet.length() - 2; lastWordBreak >= 0; --lastWordBreak)
										   if (lastPlainTextSet.at(lastWordBreak).isSpace())
											   break;
									   QString lastWord = lastPlainTextSet.mid(lastWordBreak + 1, -2);
									   VuoText lastWordAsVuoText = VuoText_make(lastWord.toUtf8().constData());
									   typedWord(lastWordAsVuoText);
								   }

								   if (currChar == QString('\n'))
								   {
									   // Fire an event for each line added in the most recent edit.
									   int lastLineBreak = lastPlainTextSet.lastIndexOf('\n');
									   QString lastLine = lastPlainTextSet.mid(lastLineBreak + 1, -2);
									   VuoText lastLineAsVuoText = VuoText_make(lastLine.toUtf8().constData());
									   typedLine(lastLineAsVuoText);
								   }
							   }

							   lastPlainTextSet.append(currChar);
						   }
					   }
				   });
}

/**
 * Moves the cursor to the end of the text edit widget.
 *
 * Must be called on the main thread.
 */
void VuoWindowTextInternal::moveCursorToEnd(void)
{
	if (! pte->textCursor().atEnd())
		pte->moveCursor(QTextCursor::End);
}

/**
 * Prevents the window from closing.
 */
void VuoWindowTextInternal::closeEvent(QCloseEvent *event)
{
	event->ignore();
}
