/**
 * @file
 * vuo.device.mouse node implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"

#include <ApplicationServices/ApplicationServices.h>

const char *moduleName = "Get Mouse";
const char *moduleDescription = "Gets the current state of the mouse, and generates events when the mouse is clicked or moved.  `moved` provides the delta from the last pointer position (even if the pointer is already on the edge of the screen and moves further off the screen).  `scrolled` provides the delta from the last scroll position.  `position` is in screen coordinates (use the @@@ node to transform to window coordinates, or the @@@ to transform to scene coordinates.";
const unsigned int moduleVersion = 100;
const bool nodeIsInterface = true;
const char *moduleKeywords[] = { "@@@" };

struct nodeInstanceData
{
	CFMachPortRef tap;
	CFRunLoopSourceRef source;
	void (*moved)(VuoPoint2d);
	// void (*leftPressed)(VuoPoint2d); ...
};

CGEventRef vuoDeviceMouseCGEventTapCallback(CGEventTapProxy proxy, CGEventType type, CGEventRef event, void *refcon)
{
	struct nodeInstanceData * context = (struct nodeInstanceData *)refcon;

	if (CGEventGetType(event) == kCGEventMouseMoved)
	{
		VuoPoint2d point = { CGEventGetDoubleValueField(event,kCGMouseEventDeltaX), CGEventGetDoubleValueField(event,kCGMouseEventDeltaY) };
		context->moved(point);
	}

	return NULL; // Return value is ignored since we're kCGEventTapOptionListenOnly.
}

struct nodeInstanceData * nodeInstanceInit(void)
{
	struct nodeInstanceData *context = (struct nodeInstanceData *)calloc(1,sizeof(struct nodeInstanceData));
	vuoRegister(context, free);
	return context;
}

void nodeInstanceCallbackStart
(
		VuoInstanceData(struct nodeInstanceData *) context,
		VuoOutputTrigger(moved, VuoPoint2d)
		//	VuoOutputTrigger(leftPressed, VuoPoint2d),
		//	VuoOutputTrigger(leftReleased, VuoPoint2d),
		//	VuoOutputTrigger(middlePressed, VuoPoint2d),
		//	VuoOutputTrigger(middleReleased, VuoPoint2d),
		//	VuoOutputTrigger(rightPressed, VuoPoint2d),
		//	VuoOutputTrigger(rightReleased, VuoPoint2d),
		//	VuoOutputTrigger(scrolled, VuoReal)
)
{
	(*context)->moved = moved;
	// context->leftPressed = ...

	(*context)->tap = CGEventTapCreate(kCGSessionEventTap, kCGHeadInsertEventTap, kCGEventTapOptionListenOnly, kCGEventMaskForAllEvents, vuoDeviceMouseCGEventTapCallback, *context);
	CGEventTapEnable((*context)->tap, true);
	(*context)->source = CFMachPortCreateRunLoopSource(kCFAllocatorDefault, (*context)->tap, 0);
	CFRunLoopAddSource(CFRunLoopGetMain(), (*context)->source, kCFRunLoopCommonModes);
}

void nodeInstanceEvent
(
		VuoInstanceData(struct nodeInstanceData *) context,
		VuoOutputData(VuoPoint2d) position,
		VuoOutputData(VuoInteger) isLeftPressed, /// @todo change to bool after https://b33p.net/kosada/node/4016
		VuoOutputData(VuoInteger) isMiddlePressed, /// @todo change to bool after https://b33p.net/kosada/node/4016
		VuoOutputData(VuoInteger) isRightPressed /// @todo change to bool after https://b33p.net/kosada/node/4016
)
{
	CGEventRef event = CGEventCreate(NULL);
	CGPoint p = CGEventGetLocation(event);
	CFRelease(event);
	position->x = p.x;
	position->y = p.y;

	*isLeftPressed = CGEventSourceButtonState(kCGEventSourceStateCombinedSessionState, kCGMouseButtonLeft);
	*isMiddlePressed = CGEventSourceButtonState(kCGEventSourceStateCombinedSessionState, kCGMouseButtonCenter);
	*isRightPressed = CGEventSourceButtonState(kCGEventSourceStateCombinedSessionState, kCGMouseButtonRight);
}

void nodeInstanceCallbackStop
(
		VuoInstanceData(struct nodeInstanceData *) context
)
{
	CFRunLoopRemoveSource(CFRunLoopGetMain(), (*context)->source, kCFRunLoopCommonModes);
	CFRelease((*context)->source);
	CGEventTapEnable((*context)->tap, false);
	CFRelease((*context)->tap);
}

void nodeInstanceFini
(
		VuoInstanceData(struct nodeInstanceData *) context
)
{
}
