/**
 * @file
 * vuo.image.render.window node implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"
#include "VuoWindow.h"
#include "VuoSceneRenderer.h"

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <dispatch/dispatch.h>

#include <OpenGL/gl.h>
//#import <OpenGL/gl3.h>
/// @todo After we drop 10.6 support, switch back to gl3 and remove the below 4 lines.  See also r15430 for shader changes.
#include <OpenGL/glext.h>


const char *moduleName = "Render Image to Window";
const char *moduleDescription = "Creates a window, and displays an image on it.";
const unsigned int moduleVersion = 100;
const bool nodeIsInterface = true;
const char *moduleKeywords[] = { "draw", "graphics" };
/// @todo remove Qt dependencies after https://b33p.net/kosada/node/4018 is fixed
const char *moduleDependencies[] = {
	"VuoGlContext",
	"VuoSceneRenderer.bc",
	"VuoWindow.bc",
	"VuoWindowApplication.bc",
	"VuoWindowOpenGLInternal.bc",
	"VuoWindowTextInternal.bc",
	"moc_VuoWindowApplication.bc",
	"moc_VuoWindowOpenGLInternal.bc",
	"moc_VuoWindowTextInternal.bc",
	"QtCore.framework",
	"QtGui.framework",
	"QtWidgets.framework",
	"QtOpenGL.framework",
	"objc",
	"z",
	"CoreFoundation.framework",
	"AppKit.framework",
	"Carbon.framework",
	"Security.framework"
};


struct nodeInstanceData
{
	VuoWindowOpenGL *window;
	VuoSceneRenderer sceneRenderer;

	dispatch_semaphore_t scenegraphSemaphore; ///< Serializes access to @c rootSceneObject.
	VuoSceneObject rootSceneObject;
};

void vuo_image_render_window_init(void *ctx)
{
	struct nodeInstanceData *context = ctx;

	VuoSceneRenderer_prepareContext(context->sceneRenderer);
}

void vuo_image_render_window_resize(void *ctx, unsigned int width, unsigned int height)
{
	struct nodeInstanceData *context = ctx;
	VuoSceneRenderer_regenerateProjectionMatrix(context->sceneRenderer, width, height);
}

void vuo_image_render_window_draw(void *ctx)
{
	struct nodeInstanceData *context = ctx;

	glClearColor(0,0,0,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	dispatch_semaphore_wait(context->scenegraphSemaphore, DISPATCH_TIME_FOREVER);
	VuoSceneRenderer_draw(context->sceneRenderer);
	dispatch_semaphore_signal(context->scenegraphSemaphore);
}

struct nodeInstanceData *nodeInstanceInit(void)
{
	struct nodeInstanceData *context = (struct nodeInstanceData *)calloc(1,sizeof(struct nodeInstanceData));

	context->sceneRenderer = VuoSceneRenderer_create();

	VuoList_VuoVertices verticesList = VuoListCreate_VuoVertices();
	VuoListAppendValue_VuoVertices(verticesList, VuoVertices_getQuad());
	context->rootSceneObject = VuoSceneObject_make(
				verticesList,
				VuoShader_makeImageShader(),
				VuoTransform_makeIdentity(),
				VuoListCreate_VuoSceneObject()
			);
	context->rootSceneObject.transform.scale.x = 2;
	VuoSceneRenderer_setRootSceneObject(context->sceneRenderer, context->rootSceneObject);

	context->scenegraphSemaphore = dispatch_semaphore_create(1);

	context->window = VuoWindowOpenGLCreate(
				vuo_image_render_window_init,
				vuo_image_render_window_resize,
				vuo_image_render_window_draw,
				(void *)context
			);

	vuoRegister(context, free);
	return context;
}

void nodeInstanceCallbackStart
(
		VuoInstanceData(struct nodeInstanceData *) context
)
{
	VuoWindowOpenGLEnableTriggers((*context)->window, NULL, NULL, NULL);
}

void nodeInstanceEvent
(
		VuoInstanceData(struct nodeInstanceData *) context,
		VuoInputData(VuoImage) image
)
{
	dispatch_semaphore_wait((*context)->scenegraphSemaphore, DISPATCH_TIME_FOREVER);
	{
		VuoShader_resetTextures((*context)->rootSceneObject.shader);
		if (image)
		{
			VuoShader_addTexture((*context)->rootSceneObject.shader, image, "texture");
			(*context)->rootSceneObject.transform.scale.y = 2. * (float)image->pixelsHigh/(float)image->pixelsWide;
			VuoWindowOpenGLSetAspectRatio((*context)->window, image->pixelsWide, image->pixelsHigh);
		}

		VuoSceneRenderer_setRootSceneObject((*context)->sceneRenderer, (*context)->rootSceneObject);
	}
	dispatch_semaphore_signal((*context)->scenegraphSemaphore);

	// Schedule a redraw.
	VuoWindowOpenGLRedraw((*context)->window);
}

void nodeInstanceCallbackStop
(
		VuoInstanceData(struct nodeInstanceData *) context
)
{
	VuoWindowOpenGLDisableTriggers((*context)->window);
}

void nodeInstanceFini
(
		VuoInstanceData(struct nodeInstanceData *) context
)
{
	VuoWindowOpenGLDestroy((*context)->window);
	VuoSceneRenderer_destroy((*context)->sceneRenderer);
	dispatch_release((*context)->scenegraphSemaphore);
}
