/**
 * @file
 * vuo.scene.render.window node implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"
#include "VuoWindow.h"
#include "VuoSceneRenderer.h"

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <OpenGL/gl.h>
//#import <OpenGL/gl3.h>
/// @todo After we drop 10.6 support, switch back to gl3 and remove the below 4 lines.  See also r15430 for shader changes.
#include <OpenGL/glext.h>

#define DEBUG 0

const char *moduleName = "Render Scene to Window";
const char *moduleDescription = "Renders `rootSceneObject` and its descendants.  When the display is ready for the next frame, `frameRequested` is fired.  When the user moves the mouse inside the window, `mouseMoved` is fired with the current position of the mouse, in root scene coordinates.  (For additional mouse functionality, see `vuo.device.mouse`.)";
const unsigned int moduleVersion = 100;
const bool nodeIsInterface = true;
const char *moduleKeywords[] = { "draw", "opengl", "scenegraph", "graphics" };
/// @todo remove Qt dependencies after https://b33p.net/kosada/node/4018 is fixed
const char *moduleDependencies[] = {
	"VuoGlContext",
	"VuoSceneRenderer.bc",
	"VuoWindow.bc",
	"VuoWindowApplication.bc",
	"VuoWindowOpenGLInternal.bc",
	"VuoWindowTextInternal.bc",
	"moc_VuoWindowApplication.bc",
	"moc_VuoWindowOpenGLInternal.bc",
	"moc_VuoWindowTextInternal.bc",
	"QtCore.framework",
	"QtGui.framework",
	"QtWidgets.framework",
	"QtOpenGL.framework",
	"objc",
	"z",
	"CoreFoundation.framework",
	"AppKit.framework",
	"Carbon.framework",
	"Security.framework"
};


struct nodeInstanceData
{
	VuoWindowOpenGL *window;
	VuoSceneRenderer *sceneRenderer;
};

void vuo_scene_render_window_init(void *ctx)
{
	struct nodeInstanceData *context = ctx;

	VuoSceneRenderer_prepareContext(context->sceneRenderer);
}

void vuo_scene_render_window_resize(void *ctx, unsigned int width, unsigned int height)
{
	//fprintf(stderr, "vuo_scene_render_window_resize(%d,%d)\n", width, height);
	struct nodeInstanceData *context = ctx;

	VuoSceneRenderer_regenerateProjectionMatrix(context->sceneRenderer, width, height);
}

void vuo_scene_render_window_draw(void *ctx)
{
	//fprintf(stderr, "vuo_scene_render_window_draw\n");
	struct nodeInstanceData *context = ctx;

	glClearColor(0,0,0,1);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	VuoSceneRenderer_draw(context->sceneRenderer);

#if DEBUG
	VuoSceneRenderer_drawElement(context->sceneRenderer, 0, .08f);	// Normals
	VuoSceneRenderer_drawElement(context->sceneRenderer, 1, .08f);	// Tangents
	VuoSceneRenderer_drawElement(context->sceneRenderer, 2, .08f);	// Bitangents
#endif
}

struct nodeInstanceData *nodeInstanceInit(void)
{
	struct nodeInstanceData *context = (struct nodeInstanceData *)calloc(1,sizeof(struct nodeInstanceData));
	context->sceneRenderer = VuoSceneRenderer_create();
	context->window = VuoWindowOpenGLCreate(
				vuo_scene_render_window_init,
				vuo_scene_render_window_resize,
				vuo_scene_render_window_draw,
				(void *)context
			);
	vuoRegister(context, free);
	return context;
}

void nodeInstanceCallbackStart
(
		VuoInstanceData(struct nodeInstanceData *) context,
		VuoOutputTrigger(frameRequested, VuoInteger), /// @todo change to a struct containing timestamp and frameCount
		VuoOutputTrigger(mouseMoved, VuoInteger) /// @todo change to VuoPoint2d

		//	VuoOutputTrigger(leftPressed,VuoPoint2d),
		//	VuoOutputTrigger(leftReleased,VuoPoint2d),
		//	VuoOutputTrigger(middlePressed,VuoPoint2d),
		//	VuoOutputTrigger(middleReleased,VuoPoint2d),
		//	VuoOutputTrigger(rightPressed,VuoPoint2d),
		//	VuoOutputTrigger(rightReleased,VuoPoint2d)
		//	VuoOutputTrigger(scrolled,VuoReal)

)
{
	VuoWindowOpenGLEnableTriggers((*context)->window, frameRequested, NULL, NULL);  /// @todo
}

void nodeInstanceEvent
(
		VuoInstanceData(struct nodeInstanceData *) context,
		VuoInputData(VuoSceneObject) rootSceneObject
)
{
	VuoSceneRenderer_setRootSceneObject((*context)->sceneRenderer, rootSceneObject);

	// Schedule a redraw.
	VuoWindowOpenGLRedraw((*context)->window);
}

void nodeInstanceCallbackStop
(
		VuoInstanceData(struct nodeInstanceData *) context
)
{
	VuoWindowOpenGLDisableTriggers((*context)->window);
}

void nodeInstanceFini
(
		VuoInstanceData(struct nodeInstanceData *) context
)
{
	VuoWindowOpenGLDestroy((*context)->window);
	VuoSceneRenderer_destroy((*context)->sceneRenderer);
}
