/**
 * @file
 * VuoRendererColors implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code is proprietary, and may not be redistributed.
 * For more information, see http://vuo.org/license.
 */

#include "VuoRendererColors.hh"

const qreal VuoRendererColors::minNodeFrameAndFillAlpha = 0.35;
const qreal VuoRendererColors::maxNodeFrameAndFillAlpha = 1.00;
const qreal VuoRendererColors::defaultNodeFrameAndFillAlpha = 0.75;
const qreal VuoRendererColors::defaultCableUpperAndMainAlpha = 0.5;
const qreal VuoRendererColors::defaultPortFillAlpha = 3./8.;
const int VuoRendererColors::activityFadeDuration = 400;
const int VuoRendererColors::activityAnimationFadeDuration = 950;

/**
 * Creates a new color scheme provider, optionally tinted with @c tintColor.
 * If @c isSelected is true, the colors are also tinted dark blue to indicate selection.
 * If @c isHovered is true, the colors are also slightly tinted dark blue to indicate potential for selection.
 * If @c isHighlighed is true, the colors are also tinted with light blue (more easily visible at a distance) to indicate potential for cable connection.
 * If @c timeOfLastActivity is anything other than VuoRendererItem::notTrackingActivity, the alpha level
 * is modified to reflect the amount of time that has passed since the @c timeOfLastActivity (e.g., a node execution or event firing), in ms since epoch.
 */
VuoRendererColors::VuoRendererColors(VuoNode::TintColor tintColor, bool isSelected, bool isHovered, bool isHighlighted, qint64 timeOfLastActivity)
{
	this->tintColor = tintColor;
	this->isSelected = isSelected;
	this->isHovered = isHovered;
	this->isHighlighted = isHighlighted;

	// Turn composition components opaque as they execute or fire events, then fade them back to their minimum transparency.
	qint64 timeNow = QDateTime::currentMSecsSinceEpoch();
	const qreal defaultFadePercentage = (maxNodeFrameAndFillAlpha-defaultNodeFrameAndFillAlpha)/(maxNodeFrameAndFillAlpha-minNodeFrameAndFillAlpha);

	this->currentFadePercentage =	((timeOfLastActivity == VuoRendererItem::notTrackingActivity)? defaultFadePercentage :
									((timeOfLastActivity == VuoRendererItem::activityInProgress)? 0 :
									fmin(1, (timeNow - timeOfLastActivity)/(1.0*activityFadeDuration))));
}

VuoRendererColors *VuoRendererColors::sharedColors = NULL;

/**
 * Returns a shared, untinted color scheme provider.
 * Don't free this object.
 */
VuoRendererColors *VuoRendererColors::getSharedColors(void)
{
	if (!sharedColors)
		sharedColors = new VuoRendererColors();

	return sharedColors;
}

/**
 * Returns the color for the background of the composition canvas.
 * Can be overridden by @c VuoRendererComposition::setBackgroundTransparent.
 */
QColor VuoRendererColors::canvasFill(void)
{
	return QColor::fromHslF(0, 0, 1, 1);
}

/**
 * Returns the color for the background of the main section of the node (the background of the area the port labels are drawn on).
 * Also used for the background of collapsed typecast ports.
 */
QColor VuoRendererColors::nodeFill(void)
{
	qreal adjustedAlpha = getCurrentAlphaForDefault(defaultNodeFrameAndFillAlpha);
	QColor nodeFillColor(tint(QColor::fromHslF(0, 0, 3./4., adjustedAlpha)));

	return nodeFillColor;
}

/**
 * Returns the color for the node's outline and the background of its title bar.
 * Also used for collapsed typecast port borders.
 */
QColor VuoRendererColors::nodeFrame(void)
{
	qreal adjustedAlpha = getCurrentAlphaForDefault(defaultNodeFrameAndFillAlpha);
	QColor nodeFrameColor(tint(QColor::fromHslF(0, 0, 1./2., adjustedAlpha)));

	return nodeFrameColor;
}

/**
 * Returns the color for the node's title text in the node's title bar.
 */
QColor VuoRendererColors::nodeTitle(void)
{
	return tint(QColor::fromHslF(0, 0, 1, 7./8.));
}

/**
 * Returns the color for the node's class name text in the node's title bar.
 */
QColor VuoRendererColors::nodeClass(void)
{
	return tint(QColor::fromHslF(0, 0, 1, .5));
}

/**
 * Returns the color for the background of ports in the main section of the node.
 */
QColor VuoRendererColors::portFill(void)
{
	qreal adjustedAlpha = getCurrentAlphaForDefault(defaultPortFillAlpha);
	return tint(QColor::fromHslF(0, 0, 3./4., adjustedAlpha));
}

/**
 * Returns the color for the background of an animated copy of a port.
 */
QColor VuoRendererColors::animatedPortFill(void)
{
	qreal minAlpha = 0;
	qreal maxAlpha = getMaxAlphaForDefault(defaultPortFillAlpha);
	qreal adjustedAlpha = maxAlpha-currentFadePercentage*(maxAlpha-minAlpha);

	return tint(QColor::fromHslF(0, 0, 3./4., adjustedAlpha));
}

/**
 * Returns the color for the background of ports in the titlebar of the node.
 */
QColor VuoRendererColors::portTitlebarFill(void)
{
	return nodeFrame();
}

/**
 * Returns the color for a port's conductivity barrier.
 */
QColor VuoRendererColors::conductivityBarrier(void)
{
	return nodeFrame();
}

/**
 * Returns the color for the conductivity barrier of the animated copy of a port.
 */
QColor VuoRendererColors::animatedConductivityBarrier(void)
{
	qreal minAlpha = 0;
	qreal maxAlpha = getMaxAlphaForDefault(defaultNodeFrameAndFillAlpha);
	qreal adjustedAlpha = maxAlpha-currentFadePercentage*(maxAlpha-minAlpha);

	return tint(QColor::fromHslF(0, 0, 1./2., adjustedAlpha));
}

/**
 * Returns the color for the title text of ports in the main section of the node.
 * Also used for constant value flag text.
 */
QColor VuoRendererColors::portTitle(void)
{
	const qreal defaultPortTitleAlpha = 1.;
	qreal adjustedAlpha = getCurrentAlphaForDefault(defaultPortTitleAlpha);

	return tint(QColor::fromHslF(0, 0, 1./4., adjustedAlpha));
}

/**
 * Returns the color for the highlight on the upper side of left-to-right cables (and lower side of right-to-left cables).
 */
QColor VuoRendererColors::cableUpper(void)
{
	qreal adjustedAlpha = getCurrentAlphaForDefault(defaultCableUpperAndMainAlpha);
	return tint(QColor::fromHslF(0, 0, .75, adjustedAlpha), 2.);
}

/**
 * Returns the color for the main cable body.
 */
QColor VuoRendererColors::cableMain(void)
{
	qreal adjustedAlpha = getCurrentAlphaForDefault(defaultCableUpperAndMainAlpha);
	return tint(QColor::fromHslF(0, 0, .25, adjustedAlpha), 2.);
}

/**
 * Returns the linear interpolation in RGB colorspace at position @c t (0..1) between colors @c v0 and @c v1.
 * Uses alpha channel from @c v0.
 */
QColor VuoRendererColors::lerpColor(QColor v0, QColor v1, float t)
{
	qreal r0,g0,b0,a;
	v0.getRgbF(&r0,&g0,&b0,&a);

	qreal r1,g1,b1;
	v1.getRgbF(&r1,&g1,&b1);

	return QColor::fromRgbF(
				fmin(r0 + (r1 - r0)*t, 1.),
				fmin(g0 + (g1 - g0)*t, 1.),
				fmin(b0 + (b1 - b0)*t, 1.),
				a
				);
}

/**
 * Tints the specified color by @c tintColor.
 * @c amount is used to tint cables more strongly, since their small size makes the tint less apparent.
 */
QColor VuoRendererColors::tint(QColor color, qreal amount)
{
	qreal h,s,l,a;
	color.getHslF(&h,&s,&l,&a);

	qreal hs;
	if (tintColor == VuoNode::TintYellow)
		hs = 45./360.;
	else if (tintColor == VuoNode::TintOrange)
		hs = 15./360.;
	else if (tintColor == VuoNode::TintMagenta)
		hs = 315./360.;
	else if (tintColor == VuoNode::TintViolet)
		hs = 255./360.;
	else if (tintColor == VuoNode::TintCyan)
		hs = 175./360.;
	else if (tintColor == VuoNode::TintGreen)
		hs = 90./360.;

	if (tintColor != VuoNode::TintNone)
	{
		// Factor in the source hue only if source saturation is nonzero
		if (s > .1)
			h = h + (hs - h) * 3./4.;
		else
			h = hs;

		s = s + (1. - s) * 3./8. * amount;
	}

	color = QColor::fromHslF(h,s,l,a);

	if (isHighlighted)
	{
		QColor highlight = QColor::fromHslF(235./360., 1., .7, 1.); // Light Blue
		color = lerpColor(color, highlight, 7./8. * amount);
	}

	if (isSelected || isHovered)
	{
		QColor selection = QColor::fromHslF(235./360., 1., .3, 1.); // Dark Blue
		color = lerpColor(color, selection, (isSelected * 4./8. + isHovered * 3./8.) * amount);
	}

	return color;
}

/**
 * Returns the projected time (in ms since epoch) at which a node must have completed its
 * previous execution in order to have faded to the @c defaultNodeFrameAndFillAlpha level
 * at the current time.
 */
qint64 VuoRendererColors::getVirtualNodeExecutionOrigin(void)
{
	const qreal defaultFadePercentage = (maxNodeFrameAndFillAlpha-defaultNodeFrameAndFillAlpha)/(maxNodeFrameAndFillAlpha-minNodeFrameAndFillAlpha);
	qint64 timeNow = QDateTime::currentMSecsSinceEpoch();
	qint64 virtualNodeExecutionOrigin = timeNow - defaultFadePercentage*activityFadeDuration;

	return virtualNodeExecutionOrigin;
}

/**
 * Returns the projected time (in ms since epoch) at which a port must have fired its
 * most recent event in order to have faded to the @c defaultPortFillAlpha level
 * at the current time.
 */
qint64 VuoRendererColors::getVirtualFiredEventOrigin(void)
{
	qreal minPortFillAlpha = getMinAlphaForDefault(defaultPortFillAlpha);
	qreal maxPortFillAlpha = getMaxAlphaForDefault(defaultPortFillAlpha);

	const qreal defaultFadePercentage = (maxPortFillAlpha-defaultPortFillAlpha)/(maxPortFillAlpha-minPortFillAlpha);
	qint64 timeNow = QDateTime::currentMSecsSinceEpoch();
	qint64 virtualFiredEventOrigin = timeNow - defaultFadePercentage*activityFadeDuration;

	return virtualFiredEventOrigin;
}

/**
 * Returns the projected time (in ms since epoch) at which a port must have fired its
 * most recent event in order for its animation to have faded to the @c percentage provided.
 */
qint64 VuoRendererColors::getVirtualFiredEventOriginForAnimationFadePercentage(qreal percentage)
{
	qint64 timeNow = QDateTime::currentMSecsSinceEpoch();
	qint64 virtualFiredEventOrigin = timeNow - percentage*activityAnimationFadeDuration;

	return virtualFiredEventOrigin;
}

/**
 * Returns the projected time (in ms since epoch) at which a cable must have propagated its
 * most recent event in order to have faded to the @c defaultCableUpperAndMainAlpha level
 * at the current time.
 */
qint64 VuoRendererColors::getVirtualPropagatedEventOrigin(void)
{
	qreal minCableUpperAndMainAlpha = getMinAlphaForDefault(defaultCableUpperAndMainAlpha);
	qreal maxCableUpperAndMainAlpha = getMaxAlphaForDefault(defaultCableUpperAndMainAlpha);

	const qreal defaultFadePercentage = (maxCableUpperAndMainAlpha-defaultCableUpperAndMainAlpha)/(maxCableUpperAndMainAlpha-minCableUpperAndMainAlpha);
	qint64 timeNow = QDateTime::currentMSecsSinceEpoch();
	qint64 virtualFiredEventOrigin = timeNow - defaultFadePercentage*activityFadeDuration;

	return virtualFiredEventOrigin;
}

/**
 * Makes use of the primary {min,max,default}NodeFrameAndFillAlpha values to calculate a proportional
 * alpha range for the provided @c defaultAlpha level; based on that range and the
 * @c currentFadePercentage, returns the appropriate current alpha level.
 */
qreal VuoRendererColors::getCurrentAlphaForDefault(qreal defaultAlpha)
{
	qreal minAlpha = getMinAlphaForDefault(defaultAlpha);
	qreal maxAlpha = getMaxAlphaForDefault(defaultAlpha);
	qreal adjustedAlpha = maxAlpha-currentFadePercentage*(maxAlpha-minAlpha);

	return adjustedAlpha;
}

/**
 * Makes use of the primary {min,default}NodeFrameAndFillAlpha values to calculate a proportional
 * minimum alpha value for the provided default alpha level.
 */
qreal VuoRendererColors::getMinAlphaForDefault(qreal defaultAlpha)
{
	qreal primaryAlphaMinToDefaultRatio = minNodeFrameAndFillAlpha/defaultNodeFrameAndFillAlpha;
	return primaryAlphaMinToDefaultRatio*defaultAlpha;
}

/**
 * Makes use of the primary {default,max}NodeFrameAndFillAlpha values to calculate a proportional
 * maximum alpha value for the provided default alpha level.
 */
qreal VuoRendererColors::getMaxAlphaForDefault(qreal defaultAlpha)
{
	qreal primaryAlphaMaxToDefaultRatio = maxNodeFrameAndFillAlpha/defaultNodeFrameAndFillAlpha;
	return fmin(1.0, primaryAlphaMaxToDefaultRatio*defaultAlpha);
}
