﻿/**
 * @file
 * VuoRendererNode interface.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This interface description may be modified and distributed under the terms of the GNU Lesser General Public License (LGPL) version 2 or later.
 * For more information, see http://vuo.org/license.
 */

#ifndef VUORENDERERNODE_HH
#define VUORENDERERNODE_HH

#include "VuoRendererItem.hh"
#include "VuoRendererPortList.hh"
#include "VuoRendererSignaler.hh"

#include "VuoBaseDetail.hh"
#include "VuoNode.hh"

class VuoCompilerNode;
class VuoCompilerNodeClass;
class VuoRendererColors;
class VuoRendererPort;
class VuoNodeClass;
class VuoCable;

/**
 * Renders a node in a @c VuoRendererComposition.
 */
class VuoRendererNode : public VuoRendererItem, public VuoBaseDetail<VuoNode>
{
public:
	/**
	 * Specifies how this node is graphically represented.
	 */
	enum Type
	{
		node,
		detachedConstant,
		detachedDrawer,
		detachedTypecast
	};

	VuoRendererNode(VuoNode * baseNode, VuoRendererSignaler *signaler);

	QRectF boundingRect(void) const;
	void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget);

	QRectF getNodeFrameRect(void) const;

	void setMissingImplementation(bool missingImplementation);

	void setProxyNode(VuoRendererNode * proxyNode);
	VuoRendererNode * getProxyNode(void) const;

	void setCollapsedForm(VuoRendererPort * collapsedForm);
	VuoRendererPort * getCollapsedForm(void) const;

	QVariant itemChange(GraphicsItemChange change, const QVariant &value);
	void updateGeometry(void);
	void updateConnectedCableGeometry(void);

	set<VuoCable *> getConnectedCables(bool includePublishedCables);
	set<VuoCable *> getConnectedInputCables(bool includePublishedCables);
	set<VuoCable *> getConnectedOutputCables(bool includePublishedCables);
	VuoRendererPortList * getInputPorts(void);
	VuoRendererPortList * getOutputPorts(void);
	void replaceInputPort(VuoRendererPort * oldPort, VuoRendererPort * newPort);
	void addInputPort(VuoRendererPort * newPort);
	void setDisplayName(string displayName);

	QRectF getNodeTitleBoundingRect(void) const;
	static QString generateNodeClassToolTip(VuoNodeClass *nodeClass);

	void resetTimeLastExecuted();
	void setExecutionEnded();
	void setExecutionBegun();
	qint64 getTimeLastExecutionEnded();

	// Drawing configuration
	static const qreal nodeTitleHeight;

protected:
	void hoverEnterEvent(QGraphicsSceneHoverEvent * event);
	void hoverMoveEvent(QGraphicsSceneHoverEvent * event);
	void hoverLeaveEvent(QGraphicsSceneHoverEvent * event);
	void mouseDoubleClickEvent (QGraphicsSceneMouseEvent *event);
	void keyPressEvent(QKeyEvent *event);

private:
	VuoRendererNode *proxyNode; ///< The optional node that handles the rendering for this node.
	VuoRendererPort *collapsedForm; ///< The collapsed form associated with this node, currently or ever in the past; stored for potential re-use.

	VuoRendererSignaler *signaler;

	// Drawing configuration
	static const qreal subcompositionBulge;
	static const qreal cornerRadius;
	static const qreal nodeClassHeight;

	QRectF frameRect;
	QRectF nodeTitleBoundingRect;

	// Node attributes affecting drawing
	enum Type nodeType;
	QString nodeTitle;
	QString nodeClass;
	bool nodeIsStateful;
	bool nodeIsSubcomposition;
	bool nodeIsMissing;
	QString nodeClassToolTip;
	VuoRendererPort *refreshPort;
	VuoRendererPort *functionPort;
	VuoRendererPortList *inputPorts;
	VuoRendererPortList *outputPorts;
	qint64 timeLastExecutionEnded;

	// Internal methods
	void initWithNodeClass(VuoNodeClass *nodeClass);
	void setInputPorts(vector<VuoRendererPort *> inputPorts = vector<VuoRendererPort *>());
	void setOutputPorts(vector<VuoRendererPort *> outputPorts = vector<VuoRendererPort *>());

	void layoutPorts(void);

	void drawNodeFrame(QPainter *painter, QRectF nodeInnerFrameRect, VuoRendererColors *colors) const;
	static void addRoundedCorner(QPainterPath &path, bool drawLine, QPointF sharpCornerPoint, qreal radius, bool isTop, bool isLeft);

	QPointF getPortPoint(VuoRendererPort *port, unsigned int portIndex) const;
};

#endif // VUORENDERERNODE_HH
