/**
 * @file
 * VuoHeap implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "VuoHeap.h"
#include <dispatch/dispatch.h>
#include <dlfcn.h>
#include <map>
#include <sstream>
using namespace std;

/**
 * Calls the sendError() function defined in VuoRuntime (without introducing a direct dependency on VuoRuntime).
 */
void sendErrorWrapper(const char *message)
{
	typedef void (*sendErrorType)(const char *message);
	sendErrorType sendError = (sendErrorType) dlsym(RTLD_SELF, "sendError");  // for running composition in separate process as executable or in current process
	if (! sendError)
		sendError = (sendErrorType) dlsym(RTLD_DEFAULT, "sendError");  // for running composition in separate process as dynamic libraries
	sendError(message);
}

map<const void *, int> referenceCounts;  ///< The reference count for each pointer.
map<const void *, DeallocateFunctionType> deallocateFunctions;  ///< The function to be used for deallocating each pointer.
map<const void *, string> descriptions;  ///< A human-readable description for each pointer.
dispatch_semaphore_t referenceCountsSemaphore = NULL;  ///< Synchronizes access to @c referenceCounts.

/**
 * Initializes the reference-counting system. To be called once, before any other reference-counting function calls.
 */
void initializeReferenceCounts(void)
{
	if (! referenceCountsSemaphore)
		referenceCountsSemaphore = dispatch_semaphore_create(1);
}

/**
 * Cleans up the reference-counting system. To be called once, after all other reference-counting function calls.
 */
void finalizeReferenceCounts(void)
{
	if (! referenceCounts.empty())
	{
		ostringstream errorMessage;
		errorMessage << "vuoRelease was not called enough times for:" << endl;
		for (map<const void *, int>::iterator i = referenceCounts.begin(); i != referenceCounts.end(); ++i)
		{
			const void *heapPointer = i->first;
			int referenceCount = i->second;
			string description = descriptions[heapPointer];
			errorMessage << "\t" << heapPointer << " " << description << " (reference count = " << referenceCount << ")" << endl;
		}
		sendErrorWrapper(errorMessage.str().c_str());
	}

	dispatch_release(referenceCountsSemaphore);
	referenceCountsSemaphore = NULL;
}

/**
 * Instead of this function, you probably want to use @c vuoRegister. This function is used to implement
 * the @c vuoRegister macro.
 *
 * Registers @c heapPointer to be reference-counted and stores its deallocate function
 * (unless @c heapPointer is null or is already being reference-counted).
 *
 * @param heapPointer A pointer to allocated memory on the heap.
 * @param deallocate The function to be used to deallocate the memory when the reference count gets back to its original value of 0.
 * @param file The name of the file in which @c vuoRegister is called.
 * @param line The line in the file in which @c vuoRegister is called.
 * @param func The function in which @c vuoRegister is called.
 * @return The updated reference count of @c heapPointer. This is 0 if @c heapPointer is not already being reference-counted, greater than 0 if it is, or -1 if @c heapPointer is null.
 */
int vuoRegisterF(const void *heapPointer, DeallocateFunctionType deallocate, const char *file, unsigned int line, const char *func)
{
	if (! heapPointer)
		return -1;

	bool isAlreadyReferenceCounted;
	int updatedCount;

	ostringstream sout;
	sout << file << ":" << line << " " << func;
	string description = sout.str();

	dispatch_semaphore_wait(referenceCountsSemaphore, DISPATCH_TIME_FOREVER);
	{

		isAlreadyReferenceCounted = (referenceCounts.find(heapPointer) != referenceCounts.end());
		if (! isAlreadyReferenceCounted)
			deallocateFunctions[heapPointer] = deallocate;
		updatedCount = referenceCounts[heapPointer];
		descriptions[heapPointer] = description;

	}
	dispatch_semaphore_signal(referenceCountsSemaphore);

	if (isAlreadyReferenceCounted)
	{
		ostringstream errorMessage;
		errorMessage << "vuoRegister was called more than once for " << heapPointer << " " << description;
		sendErrorWrapper(errorMessage.str().c_str());
	}

	return updatedCount;
}

/**
 * Increments the reference count for @c heapPointer (unless @c heapPointer is not being reference-counted).
 *
 * @param heapPointer A pointer to allocated memory on the heap.
 * @return The updated reference count of @c heapPointer, or -1 if @c heapPointer is not being reference-counted or is null.
 */
int vuoRetain(const void *heapPointer)
{
	if (! heapPointer)
		return -1;

	int updatedCount = -1;
	string description;

	dispatch_semaphore_wait(referenceCountsSemaphore, DISPATCH_TIME_FOREVER);
	{

		map<const void *, int>::iterator i = referenceCounts.find(heapPointer);
		if (i != referenceCounts.end())
			updatedCount = ++referenceCounts[heapPointer];
		description = descriptions[heapPointer];

	}
	dispatch_semaphore_signal(referenceCountsSemaphore);

	if (updatedCount == -1)
	{
		ostringstream errorMessage;
		errorMessage << "vuoRetain was called for unregistered pointer " << heapPointer << " " << description;
		sendErrorWrapper(errorMessage.str().c_str());
	}

	return updatedCount;
}

/**
 * Decrements the reference count for @c heapPointer (unless @c heapPointer is not being reference-counted).
 * If the reference count becomes 0, @c heapPointer is deallocated and is no longer reference-counted.
 *
 * @param heapPointer A pointer to allocated memory on the heap.
 * @return The updated reference count of @c heapPointer, or -1 if @c heapPointer is not being reference-counted, has never been retained, or is null.
 */
int vuoRelease(const void *heapPointer)
{
	if (! heapPointer)
		return -1;

	int updatedCount = -1;
	bool isRegisteredWithoutRetain = false;
	DeallocateFunctionType deallocate = NULL;
	string description;

	dispatch_semaphore_wait(referenceCountsSemaphore, DISPATCH_TIME_FOREVER);
	{

		map<const void *, int>::iterator i = referenceCounts.find(heapPointer);
		if (i != referenceCounts.end())
		{
			if (referenceCounts[heapPointer] == 0)
			{
				isRegisteredWithoutRetain = true;
			}
			else
			{
				updatedCount = --referenceCounts[heapPointer];

				if (updatedCount == 0)
				{
					referenceCounts.erase(heapPointer);
					deallocate = deallocateFunctions[heapPointer];
					deallocateFunctions.erase(heapPointer);
				}
			}
		}
		description = descriptions[heapPointer];

	}
	dispatch_semaphore_signal(referenceCountsSemaphore);

	if (updatedCount == 0)
	{
		deallocate((void *)heapPointer);
	}
	else if (updatedCount == -1)
	{
		ostringstream errorMessage;
		errorMessage << "vuoRelease was called for "
					 << (isRegisteredWithoutRetain ? "unretained" : "unregistered")
					 << " pointer " << heapPointer << " " << description;
		sendErrorWrapper(errorMessage.str().c_str());
	}

	return updatedCount;
}
