/**
 * @file
 * VuoHeap interface.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#ifndef VUOHEAP_H
#define VUOHEAP_H

#include <stdlib.h>
#include <stdbool.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * A type for destructor functions, such as @c free(void *), which are used to deallocate reference-counted
 * memory when it's no longer in use.
 */
typedef void (*DeallocateFunctionType)(void *);

void initializeReferenceCounts(void);
void finalizeReferenceCounts(void);

/**
 * Registers @c heapPointer to be reference-counted and stores its deallocate function
 * (unless @c heapPointer is null or is already being reference-counted).
 *
 * @param heapPointer A pointer to allocated memory on the heap.
 * @param deallocate The function to be used to deallocate the memory when the reference count gets back to its original value of 0.
 * @return The updated reference count of @c heapPointer. This is 0 if @c heapPointer is not already being reference-counted, greater than 0 if it is, or -1 if @c heapPointer is null.
 */
#define vuoRegister(heapPointer, deallocate) vuoRegisterF(heapPointer, deallocate, __FILE__, __LINE__, __func__);

int vuoRegisterF(const void *heapPointer, DeallocateFunctionType deallocate, const char *file, unsigned int line, const char *func);
int vuoRetain(const void *heapPointer);
int vuoRelease(const void *heapPointer);

#ifdef __cplusplus
}
#endif

#endif
