﻿/**
 * @file
 * TestVuoCompilerNode interface and implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code is proprietary, and may not be redistributed.
 * For more information, see http://vuo.org/license.
 */

#include <libgen.h>
#include <fcntl.h>
#include "TestVuoCompiler.hh"
#include "VuoCompiler.hh"
#include "VuoCompilerNode.hh"
#include "VuoCompilerTriggerPort.hh"
#include "VuoCompilerOutputEventPort.hh"
#include "VuoPort.hh"
#include "../../type/VuoInteger.h"


class TestVuoCompilerNode;
typedef Module * (TestVuoCompilerNode::*moduleFunction_t)();  ///< A function that creates a @c Module.

// Be able to use these types in QTest::addColumn()
Q_DECLARE_METATYPE(moduleFunction_t);


/**
 * Tests for using nodes.
 */
class TestVuoCompilerNode : public TestVuoCompiler
{
	Q_OBJECT

private:

	Module * makeSumTestModule();
	Module * makeCounterTestModule();
	Module * makeStartTestModule();
	Module * makeTwoCountersTestModule();
	Module * makeTimerTestModule();

	int numBitsInVuoInteger(void)
	{
		return sizeof(VuoInteger) * CHAR_BIT;
	}

private slots:

	void initTestCase()
	{
		initCompiler();
	}

	void cleanupTestCase()
	{
		cleanupCompiler();
	}

	void testGlobalVariableIdentifiers()
	{
		VuoCompilerNodeClass *nodeClass = compiler->getNodeClass("vuo.math.add.integer");
		VuoCompilerNode *nodeWithDefaultName1 = nodeClass->newNode()->getCompiler();
		nodeWithDefaultName1->setGraphvizIdentifier("Add");
		VuoCompilerNode *nodeWithDefaultName2 = nodeClass->newNode()->getCompiler();
		nodeWithDefaultName2->setGraphvizIdentifier("Add2");
		VuoCompilerNode *nodeWithDefaultName3 = nodeClass->newNode()->getCompiler();
		nodeWithDefaultName3->setGraphvizIdentifier("Add3");
		VuoCompilerNode *nodeWithOverriddenName = nodeClass->newNode("Overridden")->getCompiler();

		Module module("TestGlobalVariableIdentifiers", getGlobalContext());
		nodeWithDefaultName1->generateAllocation(&module);
		nodeWithDefaultName2->generateAllocation(&module);
		nodeWithDefaultName3->generateAllocation(&module);
		nodeWithOverriddenName->generateAllocation(&module);

		QVERIFY(NULL != module.getGlobalVariable("vuo_math_add_integer__Add__runtime__refresh_event", true));
		QVERIFY(NULL != module.getGlobalVariable("vuo_math_add_integer__Add2__runtime__refresh_event", true));
		QVERIFY(NULL != module.getGlobalVariable("vuo_math_add_integer__Add3__runtime__refresh_event", true));
		QVERIFY(NULL != module.getGlobalVariable("vuo_math_add_integer__Overridden__runtime__refresh_event", true));
	}

	void testPortInitialValue_data()
	{
		QTest::addColumn< QString >("nodeClassName");
		QTest::addColumn< QString >("inputPortName");
		QTest::addColumn< QString >("expectedDefaultValue");

		QTest::newRow("non-zero integer") << "vuo.math.divide.integer" << "b" << "1";
		QTest::newRow("UTF-8 string") << "vuo.test.unicodeDefaultString" << "string" << "\"画\"";
		QTest::newRow("JSON object") << "vuo.color.get.hsl" << "color" << "{\"r\":0,\"g\":0,\"b\":0,\"a\":1}";
	}
	void testPortInitialValue()
	{
		QFETCH(QString, nodeClassName);
		QFETCH(QString, inputPortName);
		QFETCH(QString, expectedDefaultValue);

		VuoCompilerNodeClass *nodeClass = compiler->getNodeClass(nodeClassName.toStdString());
		VuoNode *node = nodeClass->newNode();
		VuoPort *basePort = node->getInputPortWithName(inputPortName.toStdString());
		VuoCompilerInputEventPort *compilerPort = dynamic_cast<VuoCompilerInputEventPort *>(basePort->getCompiler());
		QCOMPARE(QString(compilerPort->getData()->getInitialValue().c_str()), expectedDefaultValue);
	}

};

QTEST_APPLESS_MAIN(TestVuoCompilerNode)
#include "TestVuoCompilerNode.moc"
