/**
 * @file
 * VuoMIDIDevice implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "type.h"
#include "VuoMIDIDevice.h"

/// @{
#ifdef VUO_COMPILER
const char * moduleName = "MIDI Device";
const char * moduleDescription = "A set of specifications for choosing a MIDI device.";
const unsigned int moduleVersion = 100;
const char * moduleDependencies[] = { "c" };
#endif
/// @}

/**
 * @ingroup VuoMIDIDevice
 * Parses the C string @c initializer to create a new value.
 */
VuoMIDIDevice VuoMIDIDevice_valueFromString(const char * initializer)
{
	/// @todo This function body will be generated by vuo-compile (https://b33p.net/kosada/node/5521)
	json_object * js = json_tokener_parse(initializer);
	VuoMIDIDevice value = VuoMIDIDevice_valueFromJson(js);
	json_object_put(js);
	return value;
}

/**
 * @ingroup VuoMIDIDevice
 * Serializes @c value as a C string.
 */
char * VuoMIDIDevice_stringFromValue(const VuoMIDIDevice value)
{
	/// @todo This function body will be generated by vuo-compile (https://b33p.net/kosada/node/5521)
	json_object *js = VuoMIDIDevice_jsonFromValue(value);
	char *string = strdup(json_object_to_json_string_ext(js,JSON_C_TO_STRING_PLAIN));
	json_object_put(js);
	return string;
}

/**
 * @ingroup VuoMIDIDevice
 * Decodes the JSON object @c js to create a new value.
 *
 * @eg{
 *   {
 *     "id" : -1,
 *     "name" : "",
 *     "isInput" : true
 *   }
 * }
 */
VuoMIDIDevice VuoMIDIDevice_valueFromJson(json_object * js)
{
	VuoMIDIDevice md = {-1,"",true};
	json_object *o = NULL;

	if (json_object_object_get_ex(js, "id", &o))
		md.id = VuoInteger_valueFromJson(o);

	if (json_object_object_get_ex(js, "name", &o))
		md.name = VuoText_valueFromJson(o);

	if (json_object_object_get_ex(js, "isInput", &o))
		md.isInput = VuoBoolean_valueFromJson(o);

	return md;
}

/**
 * @ingroup VuoMIDIDevice
 * Encodes @c value as a JSON object.
 */
json_object * VuoMIDIDevice_jsonFromValue(const VuoMIDIDevice md)
{
	json_object *js = json_object_new_object();

	json_object *idObject = VuoInteger_jsonFromValue(md.id);
	json_object_object_add(js, "id", idObject);

	json_object *nameObject = VuoText_jsonFromValue(md.name);
	json_object_object_add(js, "name", nameObject);

	json_object *isInputObject = VuoBoolean_jsonFromValue(md.isInput);
	json_object_object_add(js, "isInput", isInputObject);

	return js;
}

/**
 * @ingroup VuoMIDIDevice
 * Returns a compact string representation of @c value (comma-separated coordinates).
 */
char * VuoMIDIDevice_summaryFromValue(const VuoMIDIDevice md)
{
	const char *inOut = md.isInput ? "input" : "output";

	if (md.id == -1 && strlen(md.name) == 0)
	{
		const char *format = "The first MIDI %s device";
		int size = snprintf(NULL,0,format,inOut);
		char *valueAsString = (char *)malloc(size+1);
		snprintf(valueAsString,size+1,format,inOut);
		return valueAsString;
	}
	else if (md.id == -1)
	{
		const char *format = "The first MIDI %s device whose name contains \"%s\"";
		int size = snprintf(NULL,0,format,inOut,md.name);
		char *valueAsString = (char *)malloc(size+1);
		snprintf(valueAsString,size+1,format,inOut,md.name);
		return valueAsString;
	}
	else if (strlen(md.name) == 0)
	{
		const char *format = "MIDI %s device #%d";
		int size = snprintf(NULL,0,format,inOut,md.id);
		char *valueAsString = (char *)malloc(size+1);
		snprintf(valueAsString,size+1,format,inOut,md.id);
		return valueAsString;
	}
	else
	{
		const char *format = "MIDI %s device #%d (\"%s\")";
		int size = snprintf(NULL,0,format,inOut,md.id,md.name);
		char *valueAsString = (char *)malloc(size+1);
		snprintf(valueAsString,size+1,format,inOut,md.id,md.name);
		return valueAsString;
	}
}

/**
 * @ingroup VuoMIDIDevice
 * Retains memory allocated for this MIDI device.
 */
void VuoMIDIDevice_retain(VuoMIDIDevice value)
{
	/// @todo This function body will be generated by vuo-compile (https://b33p.net/kosada/node/5399)
	vuoRetain((void *)value.name);
}

/**
 * @ingroup VuoMIDIDevice
 * Releases memory allocated for this MIDI device.
 */
void VuoMIDIDevice_release(VuoMIDIDevice value)
{
	/// @todo This function body will be generated by vuo-compile (https://b33p.net/kosada/node/5399)
	vuoRelease((void *)value.name);
}
