/**
 * @file
 * VuoPoint4d C type definition.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#ifndef VUOPOINT4D_H
#define VUOPOINT4D_H

#include <math.h>

/**
 * @ingroup VuoTypes
 * @defgroup VuoPoint4d VuoPoint4d
 * A floating-point 4-dimensional point.
 *
 * @{
 */

/**
 * A floating-point 4-dimensional point.
 */
typedef struct
{
	float x,y,z,w;
} VuoPoint4d;

VuoPoint4d VuoPoint4d_valueFromString(const char * initializer);
char * VuoPoint4d_stringFromValue(const VuoPoint4d value);
VuoPoint4d VuoPoint4d_valueFromJson(struct json_object * js);
struct json_object * VuoPoint4d_jsonFromValue(const VuoPoint4d value);
char * VuoPoint4d_summaryFromValue(const VuoPoint4d value);

/**
 * Returns a @c VuoPoint4d with the specified coordinates.
 */
static inline VuoPoint4d VuoPoint4d_make(float x, float y, float z, float w) __attribute__((const));
static inline VuoPoint4d VuoPoint4d_make(float x, float y, float z, float w)
{
	VuoPoint4d p = {x,y,z,w};
	return p;
}

/**
 * Returns the 3D cross-product of @c u and @c v (ignoring the @c w coordinate).
 */
static inline VuoPoint4d VuoPoint4d_crossProduct(VuoPoint4d u, VuoPoint4d v) __attribute__((const));
static inline VuoPoint4d VuoPoint4d_crossProduct(VuoPoint4d u, VuoPoint4d v)
{
	VuoPoint4d p =
	{
		u.y*v.z - u.z*v.y,
		u.z*v.x - u.x*v.z,
		u.x*v.y - u.y*v.x,
		1.
	};
	return p;
}

/**
 * Returns the dot product of @c u, @c v.
 */
static inline float VuoPoint4d_dotProduct(VuoPoint4d u, VuoPoint4d v) __attribute__((const));
static inline float VuoPoint4d_dotProduct(VuoPoint4d u, VuoPoint4d v)
{
	return u.x*v.x + u.y*v.y + u.z*v.z + u.w*v.w;
}

/**
 * Returns the 3D normalization of @c a (ignoring the @c w coordinate).
 */
static inline VuoPoint4d VuoPoint4d_normalize3d(VuoPoint4d a) __attribute__((const));
static inline VuoPoint4d VuoPoint4d_normalize3d(VuoPoint4d a)
{
	float length = sqrtf(a.x*a.x + a.y*a.y + a.z*a.z);
	VuoPoint4d p =
	{
		a.x/length,
		a.y/length,
		a.z/length,
		a.w
	};
	return p;
}

/**
 * Returns the 4D normalization of @c a.
 */
static inline VuoPoint4d VuoPoint4d_normalize4d(VuoPoint4d a) __attribute__((const));
static inline VuoPoint4d VuoPoint4d_normalize4d(VuoPoint4d a)
{
	float length = sqrtf(a.x*a.x + a.y*a.y + a.z*a.z + a.w*a.w);
	VuoPoint4d p =
	{
		a.x/length,
		a.y/length,
		a.z/length,
		a.w/length
	};
	return p;
}

/**
 * @c a + @c b.
 */
static inline VuoPoint4d VuoPoint4d_add(VuoPoint4d a, VuoPoint4d b) __attribute__((const));
static inline VuoPoint4d VuoPoint4d_add(VuoPoint4d a, VuoPoint4d b)
{
	VuoPoint4d p =
	{
		a.x + b.x,
		a.y + b.y,
		a.z + b.z,
		a.w + b.w
	};
	return p;
}

/**
 * @c a - @c b.
 */
static inline VuoPoint4d VuoPoint4d_subtract(VuoPoint4d a, VuoPoint4d b) __attribute__((const));
static inline VuoPoint4d VuoPoint4d_subtract(VuoPoint4d a, VuoPoint4d b)
{
	VuoPoint4d p =
	{
		a.x - b.x,
		a.y - b.y,
		a.z - b.z,
		a.w - b.w
	};
	return p;
}

/**
 * @c a / @c b
 */
static inline VuoPoint4d VuoPoint4d_divide(VuoPoint4d a, double b) __attribute__((const));
static inline VuoPoint4d VuoPoint4d_divide(VuoPoint4d a, double b)
{
	VuoPoint4d p =
	{
		a.x / b,
		a.y / b,
		a.z / b,
		a.w / b
	};
	return p;
}

/**
 * @c a * @c b
 */
static inline VuoPoint4d VuoPoint4d_multiply(VuoPoint4d a, double b) __attribute__((const));
static inline VuoPoint4d VuoPoint4d_multiply(VuoPoint4d a, double b)
{
	VuoPoint4d p =
	{
		a.x * b,
		a.y * b,
		a.z * b,
		a.w * b
	};
	return p;
}

/**
 * @c a * @c b
 */
static inline VuoPoint4d VuoPoint4d_multiply4d(VuoPoint4d a, VuoPoint4d b) __attribute__((const));
static inline VuoPoint4d VuoPoint4d_multiply4d(VuoPoint4d a, VuoPoint4d b)
{
	VuoPoint4d p =
	{
		a.x * b.x,
		a.y * b.y,
		a.z * b.z,
		a.w * b.w
	};
	return p;
}

/**
 * @}
 */

#endif
