/**
 * @file
 * VuoText implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "type.h"
#include "VuoText.h"
#include <string.h>


/// @{
#ifdef VUO_COMPILER
const char * moduleName = "Text";
const char * moduleDescription = "A Unicode (UTF-8) text string.";
const unsigned int moduleVersion = 100;
const char * moduleKeywords[] = { "char *", "character" };
const char * moduleDependencies[] = { "c", "icuuc", "icudata" };
#endif
/// @}

/**
 * @ingroup VuoText
 * The stpncpy function is not provided by Mac OS X implementation of <string.h>.
 * Rather than adding a dependency, just define it here.
 * This is getting replaced anyway with Unicode-aware code.
 * http://pubs.opengroup.org/onlinepubs/9699919799/functions/strncpy.html
 */
char *VuoText_stpncpy(char *dest, const char *src, size_t n)
{
	char * sp = strncpy(dest,src,n);
	if (strlen(src) < n)
		return sp + strlen(src);
	return sp + n;
}

/**
 * @ingroup VuoText
 * Parses the C string @c initializer to create a new value.
 */
VuoText VuoText_valueFromString(const char * initializer)
{
	/// @todo This function body will be generated by vuo-compile (https://b33p.net/kosada/node/5521)
	json_object * js = json_tokener_parse(initializer);
	VuoText value = VuoText_valueFromJson(js);
	json_object_put(js);
	return value;
}

/**
 * @ingroup VuoText
 * Serializes @c value as a C string.
 */
char * VuoText_stringFromValue(const VuoText value)
{
	/// @todo This function body will be generated by vuo-compile (https://b33p.net/kosada/node/5521)
	json_object *js = VuoText_jsonFromValue(value);
	char *string = strdup(json_object_to_json_string_ext(js,JSON_C_TO_STRING_PLAIN));
	json_object_put(js);
	return string;
}

/**
 * @ingroup VuoText
 * Decodes the JSON object @c js, expected to contain a string, to create a new value.
 */
VuoText VuoText_valueFromJson(json_object * js)
{
	const char *textString = "";
	if (json_object_get_type(js) == json_type_string)
		textString = json_object_get_string(js);
	return VuoText_make(textString);
}

/**
 * @ingroup VuoText
 * Encodes @c value as a JSON object.
 */
json_object * VuoText_jsonFromValue(const VuoText value)
{
	return json_object_new_string(value);
}

/**
 * @ingroup VuoText
 * Creates a new UTF-8 C string from @c value, or, if it's more than 32 characters long, creates an aposiopesis.
 *
 * @eg{Hello World!}
 * @eg{I would like to convey my greeti...}
 */
char * VuoText_summaryFromValue(const VuoText value)
{
	/// @todo handle unicode
	char * summary = (char *)malloc(32+3+1);

	char * sp = summary;
	sp = VuoText_stpncpy(sp,value,32);

	// Add ellipsis if trimmed.
	if (strlen(value) > 32)
	{
		*sp++ = '.';
		*sp++ = '.';
		*sp++ = '.';
	}

	*sp++ = 0;
	return summary;
}

/**
 * @ingroup VuoText
 * Creates a VuoText value from an unquoted string (unlike @c VuoText_valueFromString(), which expects a quoted string).
 */
VuoText VuoText_make(const char * unquotedString)
{
	VuoText text = strdup(unquotedString);
	vuoRegister(text, free);
	return text;
}

/**
 * @ingroup VuoText
 * Returns the byte offset of @c logicalCharIndex into @c text's UTF-8 representation.
 */
size_t VuoText_convertUTF8IndexToByteIndex(UText *text, int logicalCharIndex)
{
	UErrorCode status = U_ZERO_ERROR;
	UBreakIterator *bi = ubrk_open(UBRK_CHARACTER, uloc_getDefault(), NULL, 0, &status);
	ubrk_setUText(bi, text, &status);

	int curLogicalCharIndex = 0;
	int curNativeCharIndex = 0;
	while((curLogicalCharIndex < logicalCharIndex) && (curNativeCharIndex != UBRK_DONE))
	{
		curNativeCharIndex = ubrk_next(bi);
		curLogicalCharIndex++;
	}

	return curNativeCharIndex;
}

/**
 * @ingroup VuoText
 * Returns the number of UTF-8 characters in the string.
 */
size_t VuoText_length(const VuoText string)
{
	UErrorCode status = U_ZERO_ERROR;
	UText *text = utext_openUTF8(NULL, string, -1, &status);

	UBreakIterator *bi = ubrk_open(UBRK_CHARACTER, uloc_getDefault(), NULL, 0, &status);
	ubrk_setUText(bi, text, &status);

	int charCount = 0;
	while (ubrk_next(bi) != UBRK_DONE)
		++charCount;

	utext_close(text);

	return charCount;
}

/**
 * @ingroup VuoText
 * Returns the substring of @c string starting at index @c startIndex and spanning @c length UTF-8 characters.
 *
 * If @c startIndex is past the end of the string, returns the empty string.
 *
 * If @c string has fewer than @c length characters from @c startIndex to the end of @c string,
 * returns all characters from @c startIndex to the end of @c string.
 */
VuoText VuoText_substring(const VuoText string, size_t startIndex, size_t length)
{
	UErrorCode status = U_ZERO_ERROR;
	UText *text = utext_openUTF8(NULL, string, -1, &status);

	size_t startAsByteIndex = VuoText_convertUTF8IndexToByteIndex(text, startIndex);
	if (startAsByteIndex > strlen(string) - 1)
		return VuoText_make("");

	size_t endAsByteIndex = VuoText_convertUTF8IndexToByteIndex(text, startIndex + length);
	if (endAsByteIndex > strlen(string))
		endAsByteIndex = strlen(string);

	size_t lengthInBytes = endAsByteIndex - startAsByteIndex;
	if (lengthInBytes == 0)
		return VuoText_make("");

	char *substr = (char *)malloc((lengthInBytes + 1) * sizeof(char));
	vuoRegister(substr, free);
	memcpy(substr, string + startAsByteIndex, lengthInBytes);
	substr[lengthInBytes] = '\0';

	utext_close(text);

	return substr;
}

/**
 * @ingroup VuoText
 * Returns a string consisting of the elements in the @c texts array concatenated together.
 */
VuoText VuoText_append(VuoText *texts, unsigned long textsCount)
{
	// Even though the inputs are UTF-8, they're still null-terminated, so we can append them as C strings.

	size_t textsLen = 0;
	for (unsigned long i = 0; i < textsCount; ++i)
		textsLen += strlen(texts[i]);

	char *composite = (char *) malloc(textsLen + 1);

	size_t offset = 0;
	for (unsigned long i = 0; i < textsCount; ++i)
	{
		strcpy(composite + offset, texts[i]);
		offset += strlen(texts[i]);
	}
	composite[textsLen] = 0;

	VuoText compositeText = VuoText_make(composite);
	free(composite);

	return compositeText;
}
