/**
 * @file
 * Prototypes for type implementations. Replace `MyType` with the name of your type.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "VuoHeap.h"


// Headers commonly used by type implementations

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdocumentation"

	#include <unicode/utext.h>
	#include <unicode/ubrk.h>

	#include <json/json.h>

#pragma clang diagnostic pop


/**
 * @defgroup VuoTypes Built-in Types
 * Types defined by Vuo Core, for use in both Vuo Graph Language and C code.
 */



#ifdef DOXYGEN

/**
 * @addtogroup WritingTypes
 * @{
 */


/**
 * @defgroup VuoTypeMetadata Type Metadata
 * Name, version, and other metadata to be defined by types.
 *
 * @{
 */

/**
 * Default display name.  Use title case.  Typically this should be the same as the type name, but with spaces and without a namespacing prefix.
 *
 * \eg{const char * moduleName = "Scene Object";}
 */
const char * moduleName;

/**
 * A succinct summary of this type's purpose.  Use sentence capitalization, when appropriate.
 *
 * \eg{const char * moduleDescription = "A renderable 3D object.";}
 */
const char * moduleDescription;

/**
 * A list of synonyms for this type's name.
 *
 * \eg{const char * moduleKeywords[] = { "perlin", "simplex" };}
 */
const char * moduleKeywords[];

/**
 * A list of libraries or other modules this module depends upon.
 *
 * @eg{const char * moduleDependencies[] = { "c", "json-c" };}
 *
 * @see UsingLibraries
 */
const char * moduleDependencies[];

/**
 * Version, represented as an integer.  Multiply the major version by 100, then add the minor version.
 *
 * \eg{const unsigned int moduleVersion = 100; // 1.00: first public release}
 * \eg{const unsigned int moduleVersion = 101; // 1.01: minor new functionality or bugfix}
 * \eg{const unsigned int moduleVersion = 200; // 2.00: significant new functionality or significant change in input/output ports}
 */
const unsigned int moduleVersion;

/**
 * @}
 */


/**
 * @defgroup VuoTypeMethods Type Methods
 * Functions to serialize, unserialize, and summarize values of the type.
 *
 * @{
 */

/**
 * Creates a new @c MyType value by unserializing a JSON-C object.
 *
 * This function is required.
 *
 * @param js A JSON-C object from @c MyType_jsonFromValue, a constant port value defined in a composition, or a default port value defined in a node class implementation.
 * @return The unserialized value. For heap-allocated types, this function should register the value with @c vuoRegister.
 */
MyType MyType_valueFromJson(json_object *js);

/**
 * Serializes a @c MyType value to a JSON-C object.
 *
 * This function is required.
 *
 * @param value The value to serialize.
 * @return A JSON-C object. It should have the same format as the argument to @c MyType_valueFromJson.
 */
json_object * MyType_jsonFromValue(const MyType value);

/**
 * Returns a brief description of a value.
 *
 * This function is required.
 *
 * @param value The value to summarize.
 * @return The summary. It should be heap-allocated; the caller is responsible for freeing it.
 */
char * MyType_summaryFromValue(const MyType value);

/**
 * @todo https://b33p.net/kosada/node/5611
 *
 * This function is optional.
 */
char * MyType_interprocessStringFromValue(const MyType value);

/**
 * @}
 */


/**
 * @}
 */

#endif
